import java.text.DecimalFormat;
import java.util.Random;

public class Player
{
    private final double MILLI = 0.001;
    private final double maxHeadPan = Math.PI / 2;
    private final double minHeadPan = -maxHeadPan;
    private final double maxHeadTilt = Math.PI / 6;
    private final double minHeadTilt = -Math.PI / 4;
    private final double headPanSpeed = 0.5 * Math.PI;
    private final double headTiltSpeed = 0.5 * Math.PI;
    private final int maxParticles = 100;

    final double FIELD_X = 3600;
    final double OFFSET_X = 150; // Beacons set back by 15cm
    final double FIELD_Y = 5400;
    final double GOAL_WIDTH = 800;

    final double FIELD_WIDTH = 4600;
    final double FIELD_LENGTH = 6600;

    private Server server;
    public Connection connection;

    private DecimalFormat df;
    private double bodyX, bodyY, bodyT;
    private double bodyDX, bodyDY, bodyDT;
    private double bodyXBias, bodyYBias, bodyTBias;
    private double bodyXNoise, bodyYNoise, bodyTNoise;
    private double headPan, headTilt;
    private double headDesiredPan, headDesiredTilt;
    private double headPanSenseNoise, headTiltSenseNoise;
    private double[] particleX;
    private double[] particleY;

    private double estimateX;
    private double estimateY;
    private double estimateT;
    private double estimateP;
    private double estimateHeadPan;
    private double estimateHeadTilt;

    private double totalDistErr;
    private double totalAngErr;
    private int numErrCalcs;
    
    private double landmarkDBias, landmarkABias;
    private double landmarkDNoise, landmarkANoise;
    private double ballDBias, ballABias;
    private double ballDNoise, ballANoise;

    private Random random;
    private DecimalFormat format;

    public Player( Server server, Connection connection )
    {
	this.server = server;
	this.connection = connection;

	df = new DecimalFormat();
	df.setMaximumFractionDigits( 2 );
	df.setGroupingUsed( false );

	bodyX = 0;
	bodyY = 0;

	particleX = new double[ maxParticles ];
	particleY = new double[ maxParticles ];

	estimateP = 0.0;

	random = new Random();

	format = new DecimalFormat("0.0");
    }

    public synchronized void resetErr()
    {
	totalDistErr = 0.0;
	totalAngErr = 0.0;
	numErrCalcs = 0;
    }

    public synchronized void setEstimate( double x, double y, double t, double p, 
					  double pan, double tilt )
    {
	estimateX = x;
	estimateY = y;
	estimateT = t;
	estimateP = p;
	estimateHeadPan = pan;
	estimateHeadTilt = tilt;

	double diffX = bodyX - estimateX;
	double diffY = bodyY - estimateY;
	double diffT = Math.abs( normalize( bodyT - estimateT ) );
	
	double distErr = Math.sqrt( diffX * diffX +
				    diffY * diffY );
	totalDistErr += distErr;
     
	double angErr = 180.0 * diffT / Math.PI;

	totalAngErr += angErr;
	
	numErrCalcs++;
	
	System.out.println( "Inst: " +  
			    "D: " + format.format(distErr) + ", " +
			    "A: " + format.format(angErr) + " || " +
			    "Avg: " +
			    "D: " + format.format(totalDistErr/numErrCalcs) + ", " +
			    "A: " + format.format(totalAngErr/numErrCalcs) );
    }

    public synchronized double getEstimateX()
    {
	return estimateX;
    }

    public synchronized double getEstimateY()
    {
	return estimateY;
    }

    public synchronized double getEstimateT()
    {
	return estimateT;
    }

    public synchronized double getEstimateP()
    {
	return estimateP;
    }

    public synchronized double getEstimatePan()
    {
	return estimateHeadPan;
    }

    public synchronized double getEstimateTilt()
    {
	return estimateHeadTilt;
    }

    public synchronized void setParticle( int num, double x, double y, double t, double p )
    {
	particleX[ num ] = x;
	particleY[ num ] = y;
    }

    public synchronized double getParticleX( int num )
    {
	return particleX[ num ];
    }

    public synchronized double getParticleY( int num )
    {
	return particleY[ num ];
    }

    public synchronized double getX()
    {
	return bodyX;
    }

    public synchronized double getY()
    {
	return bodyY;
    }

    public synchronized double getOrientation()
    {
	return bodyT;
    }

    public synchronized double getHeadPan()
    {
	return headPan;
    }

    private double normalize( double radians )
    {
	while ( radians > Math.PI )
	    radians -= 2 * Math.PI;
	while ( radians < -Math.PI )
	    radians += 2 * Math.PI;
	return radians;
    }

    public synchronized void paramWalk( double dx, double dy, double dt )
    {
	bodyDX = dx;
	bodyDY = dy;
	bodyDT = dt;

	bodyXBias = 0;
	bodyYBias = 0;
	bodyTBias = 0;

	bodyXNoise = 0.1;
	bodyYNoise = 0.1;
	bodyTNoise = 0.1;
    }

    public synchronized void moveHead( double dp, double dt )
    {
	headDesiredPan = ( dp > maxHeadPan ) ? maxHeadPan :
	    ( dp < minHeadPan ) ? minHeadPan : dp;
	headDesiredTilt = ( dt > maxHeadTilt ) ? maxHeadTilt :
	    ( dt < minHeadTilt ) ? minHeadTilt : dt;
    }
    
    private int sign( double d )
    {
	return ( d < 0 ) ? -1 : 1;
    }

    public synchronized void updatePose( int cycleTime )
    {
	// Update body
	double dx = bodyDX * cycleTime * MILLI;
	double dy = bodyDY * cycleTime * MILLI;
	double dt = bodyDT * cycleTime * MILLI;

	double vx = dx * ( random.nextGaussian() * bodyXNoise + bodyXBias + 1 );
	double vy = dy * ( random.nextGaussian() * bodyYNoise + bodyYBias + 1 );

	//System.err.println( "orig: (" + bodyX + ", " + bodyY + ", " + Math.toDegrees( bodyT ) + ")" );
	//System.err.println( "d: (" + vx + ", " + vy + ")" );
	//System.err.println( Math.toDegrees( bodyT ) );

	double c = Math.cos( bodyT );
	double s = Math.sin( bodyT );

	bodyX += vx * c - vy * s;
	bodyY += vx * s + vy * c;
	bodyT += dt * ( random.nextGaussian() * bodyTNoise + bodyTBias + 1 );
	//System.err.println( "new: (" + bodyX + ", " + bodyY + ", " + Math.toDegrees( bodyT ) + ")" );

	bodyX = ( bodyX < -FIELD_WIDTH / 2 ) ? -FIELD_WIDTH / 2 : ( bodyX > FIELD_WIDTH / 2 ) ? FIELD_WIDTH / 2 : bodyX;
	bodyY = ( bodyY < -FIELD_LENGTH / 2 ) ? -FIELD_LENGTH / 2 : ( bodyY > FIELD_LENGTH / 2 ) ? FIELD_LENGTH / 2 : bodyY;
	bodyT = normalize( bodyT );

	
	//System.out.println( "<" + df.format( bodyX ) + ", " +
	//		    df.format( bodyY ) + ", " +
	//		    df.format( Math.toDegrees( bodyT ) ) + ">" );
	
	// Update head
	double dpan = headDesiredPan - headPan;
	double dtilt = headDesiredTilt - headTilt;

	double vpan = headPanSpeed * cycleTime * MILLI;
	double vtilt = headTiltSpeed * cycleTime * MILLI;

	if ( Math.abs( dpan ) < vpan )
	    headPan = headDesiredPan;
	else
	    headPan += vpan * sign( dpan );

	if ( Math.abs( dtilt ) < vtilt )
	    headTilt = headDesiredTilt;
	else
	    headTilt += vtilt * sign( dtilt );		
    }

    private double getLandmarkAngleFromPose( double X, double Y, 
					     double x, double y, double t )
    {
	double c = Math.cos( t );
	double s = Math.sin( t );

	double px = X*c + Y*s - x*c - y*s;
	double py = -X*s + Y*c + x*s - y*c;
	double ang = Math.atan2( py, px );

	return ang;
    }

    private double getLandmarkDistanceFromPose( double X, double Y, 
					       double x, double y )
    {
	double dx = X - x;
	double dy = Y - y;

	return Math.sqrt( dx * dx + dy * dy );
    }

    public void setPose( double x, double y, double t )
    {
	bodyX = x;
	bodyY = y;
	bodyT = t;
    }

    public void setPos( double x, double y )
    {
	bodyX = x;
	bodyY = y;
    }

    public void setAng( double t )
    {
	bodyT = normalize( t );
    }

    public synchronized void sendObservations()
    {
	final double VIEW_ANGLE = 57.6;
	final double MIN_VIEW = 90 - VIEW_ANGLE / 2;
	final double MAX_VIEW = 90 + VIEW_ANGLE / 2;

	final int LM_PINK_BLUE = 0;
	final int LM_PINK_YELLOW = 1;
	final int LM_YELLOW_GOAL_LEFT = 2;
	final int LM_YELLOW_GOAL = 3;
	final int LM_YELLOW_GOAL_RIGHT = 4;
	final int LM_YELLOW_PINK = 5;
	final int LM_BLUE_PINK = 6;
	final int LM_BLUE_GOAL_LEFT = 7;
	final int LM_BLUE_GOAL = 8;
	final int LM_BLUE_GOAL_RIGHT = 9;
	final int NUM_LANDMARKS = 10;

	double[] landmarkX = new double[ NUM_LANDMARKS ];
	double[] landmarkY = new double[ NUM_LANDMARKS ];


	landmarkX[ LM_PINK_BLUE ] = -FIELD_X / 2 - OFFSET_X;
	landmarkY[ LM_PINK_BLUE ] = -FIELD_Y / 4;
	landmarkX[ LM_PINK_YELLOW ] = -FIELD_X / 2 - OFFSET_X;
	landmarkY[ LM_PINK_YELLOW ] = FIELD_Y / 4;
	landmarkX[ LM_YELLOW_GOAL_LEFT ] = -GOAL_WIDTH / 2;
	landmarkY[ LM_YELLOW_GOAL_LEFT ] = FIELD_Y / 2;
	landmarkX[ LM_YELLOW_GOAL ] = 0;
	landmarkY[ LM_YELLOW_GOAL ] = FIELD_Y / 2;
	landmarkX[ LM_YELLOW_GOAL_RIGHT ] = GOAL_WIDTH / 2;
	landmarkY[ LM_YELLOW_GOAL_RIGHT ] = FIELD_Y / 2;
	landmarkX[ LM_YELLOW_PINK ] = FIELD_X / 2 + OFFSET_X;
	landmarkY[ LM_YELLOW_PINK ] = FIELD_Y / 4;
	landmarkX[ LM_BLUE_PINK ] = FIELD_X / 2 + OFFSET_X;
	landmarkY[ LM_BLUE_PINK ] = -FIELD_Y / 4;
	landmarkX[ LM_BLUE_GOAL_LEFT ] = GOAL_WIDTH / 2;
	landmarkY[ LM_BLUE_GOAL_LEFT ] = -FIELD_Y / 2;
	landmarkX[ LM_BLUE_GOAL ] = 0;
	landmarkY[ LM_BLUE_GOAL ] = -FIELD_Y / 2;
	landmarkX[ LM_BLUE_GOAL_RIGHT ] = -GOAL_WIDTH / 2;
	landmarkY[ LM_BLUE_GOAL_RIGHT ] = -FIELD_Y / 2;

	landmarkDBias = 0; landmarkABias = 0;
	landmarkDNoise = 0.1; landmarkANoise = 0.1;
	//landmarkDNoise = 0; landmarkANoise = 0;

	ballDBias = 0; ballABias = 0;
	ballDNoise = 0.15; ballANoise = 0.05;

	headPanSenseNoise = 0.01; headTiltSenseNoise = 0.01;

	double dp = headPan + ( random.nextGaussian() * headPanSenseNoise );
	double dt = headTilt + ( random.nextGaussian() * headTiltSenseNoise );

	// Build "sense" message
	String senseMsg = "(sense " + server.getCurrentCycle() + 
	    " (h " + (int) Math.toDegrees( dp ) +
	    " " + (int) Math.toDegrees( dt ) + "))";

	// Build "see" message
	int seenLandmarks = 0;
	String seeMsg = "(see " + server.getCurrentCycle();

	Ball b = server.getBall();
	double ba = getLandmarkAngleFromPose( b.getX(),b.getY(),
					      bodyX, bodyY,
					      bodyT + headPan );
	if ( ba >= Math.toRadians( MIN_VIEW ) && 
	     ba <= Math.toRadians( MAX_VIEW ) ) {
	    double d = getLandmarkDistanceFromPose( b.getX(),b.getY(),
						    bodyX, bodyY );
	    double dd = d * ( random.nextGaussian() * ballDNoise +
			      ballDBias + 1 );
	    double da = ba + ( random.nextGaussian() * ballANoise +
			       ballABias );
	    da = normalize( da );
		
	    seeMsg += " (b " + (int) dd + " " +
		(int) Math.toDegrees( da + headPan ) + " 85)";
	}

	for ( int i = 0; i < NUM_LANDMARKS; i++ ) {
	    double a = getLandmarkAngleFromPose( landmarkX[ i ],
						 landmarkY[ i ],
						 bodyX, bodyY,
						 bodyT + headPan );
	    if ( a >= Math.toRadians( MIN_VIEW ) && 
		 a <= Math.toRadians( MAX_VIEW ) &&
		 random.nextDouble() < 0.85 ) {
		seenLandmarks++;
		double d = getLandmarkDistanceFromPose( landmarkX[ i ],
							landmarkY[ i ],
							bodyX, bodyY );
		
		double dd = d * ( random.nextGaussian() * landmarkDNoise +
				  landmarkDBias + 1 );
		double da = a + ( random.nextGaussian() * landmarkANoise +
				  landmarkABias );
		da = normalize( da );
		
		seeMsg += " (l " + i + " " + (int) dd + " " +
		    (int) Math.toDegrees( da + headPan ) + " 90)";
	    }
	}
	seeMsg += ")";
	
	// Send messages
	connection.send( senseMsg );
	if ( seenLandmarks > 0 && !server.blinded )
	    connection.send( seeMsg );
    }

    public static void main( String[] args )
    {
	Player p = new Player( null, null );
	System.out.println( Math.toDegrees( p.getLandmarkAngleFromPose( 0,4400,200,4000,Math.toRadians(2) ) ) );
    }
}
