import java.net.*;
import java.util.*;

public class Server
{
    private int currentCycle;
    private Vector players;
    private Ball ball;
    private Monitor monitor;
    private boolean paused;
    private boolean stepped;
    public boolean blinded;

    public Server()
    {
	currentCycle = 0;
	players = new Vector();
	ball = new Ball();
	monitor = null;
	paused = false;
	stepped = false;
	blinded = false;
    }

    public void errorMsg( String msg )
    {
	if ( monitor == null )
	    System.err.println( msg );
	else
	    monitor.errorMsg( msg );
    }

    public void pause()
    {
	paused = true;
	stepped = false;
    }

    public void step()
    {
	paused = false;
	stepped = true;
    }

    public void resume()
    {
	paused = false;
    }

    public void blind()
    {
      if ( blinded ) {
	blinded = false;
      } else {
	blinded = true;
      }
    }

    public void reset()
    {
	for ( int i = 0; i < players.size(); i++ ) {
	    Player p = (Player) players.elementAt( i );
	    p.resetErr();
	}
    }

    public boolean isPaused()
    {
	return paused;
    }

    public boolean isStepped()
    {
	return stepped;
    }

    public int getCurrentCycle()
    {
	return currentCycle;
    }

    public void nextCycle()
    {
	if ( !players.isEmpty() )
	    currentCycle++;
	if ( monitor != null )
	    monitor.updateField();	
    }

    public void createPlayer( Connection connection )
    {
	Player p = new Player( this, connection );
	connection.setPlayer( p );
	players.add( p );
    }

    public Vector getPlayers()
    {
	return new Vector( players );
    }

    public Ball getBall()
    {
	return ball;
    }

    public boolean processMsg( String msg, Player player )
    {
	StringTokenizer st = 
	    new StringTokenizer( msg, "( )" );

	try {
	    String command = st.nextToken();

	    if ( command.equalsIgnoreCase( "param_walk" ) ) {
		double dx = Double.parseDouble( st.nextToken() );
		double dy = Double.parseDouble( st.nextToken() );
		double dt = Math.toRadians( Double.parseDouble( st.nextToken() ) );
		player.paramWalk( dx, dy, dt ); 
	    }
	    else if ( command.equalsIgnoreCase( "move_head" ) ) {
		//System.err.println( msg );
		double dp = Double.parseDouble( st.nextToken() );
		double dt = Math.toRadians( Double.parseDouble( st.nextToken() ) );
		player.moveHead( dp, dt );
	    }
	    else if ( command.equalsIgnoreCase( "info" ) ) {
		//System.err.println( msg );
		int numParticles = 0;
		int numBallParticles = 0;
		while ( st.hasMoreTokens() ) {
		    String obj = st.nextToken();
		    if ( obj.equalsIgnoreCase( "p" ) ) {
			double x = Double.parseDouble( st.nextToken() );
			double y = Double.parseDouble( st.nextToken() );
			double t = Math.toRadians( Double.parseDouble( st.nextToken() ) );
			double p = Double.parseDouble( st.nextToken() ) * 0.01;
			player.setParticle( numParticles++, x, y, t, p );
		    }
		    else if ( obj.equalsIgnoreCase( "b" ) ) {
			double x = Double.parseDouble( st.nextToken() );
			double y = Double.parseDouble( st.nextToken() );
			double p = Double.parseDouble( st.nextToken() ) * 0.01;
			ball.setParticle( numBallParticles++, x, y, p );
		    }
		    else if ( obj.equalsIgnoreCase( "e" ) ) {
			double x = Double.parseDouble( st.nextToken() );
			double y = Double.parseDouble( st.nextToken() );
			double t = Math.toRadians( Double.parseDouble( st.nextToken() ) );
			double p = Double.parseDouble( st.nextToken() ) * 0.01;
			double pan = Math.toRadians( Double.parseDouble( st.nextToken() ) );
			double tilt = Math.toRadians( Double.parseDouble( st.nextToken() ) );
			//System.out.println( x+","+y+","+t+","+p+","+pan+","+tilt );
			player.setEstimate( x, y, t, p, pan, tilt );
		    }
		    else {
			System.err.println( "Unknown info object: " + obj );
			return false;
		    }
		}
	    }
	    else {
		System.err.println( "Unknown command: " + command );
		return false;
	    }
	}
	catch ( Exception e ) {
	    System.err.println( "Bad message format: " + msg );
	    return false;
	}

	return true;
    }



    public Connection findConnection( InetAddress clientIP,
				      int clientPort )
    {
	for ( int i = 0; i < players.size(); i++ ) {
	    Player player = (Player) players.elementAt( i );
	    if ( player.connection.matchesClient( clientIP, clientPort ) )
		return player.connection;
	}

	return null;
    }

    public void setMonitor( Monitor monitor )
    {
	this.monitor = monitor;
    }

    public static void main( String[] args )
    {
	Server server = new Server();
	Monitor monitor = new Monitor( server );
	server.setMonitor( monitor );
	ConnectionListener cl = 
	    new ConnectionListener( server, 8675, 4096 );
	cl.start();
	Timer timer = new Timer();
	int cycleTime = 50; // milliseconds
	timer.scheduleAtFixedRate( new SimStep( server, cycleTime ), 
				   0, cycleTime );
    }
}

