#include <assert.h>

#include "Geometric.h"


/*! This function returns the sign of a give double.
  1 is positive, -1 is negative
  \param d1 first parameter
  \return the sign of this double */
int sign( double d1 )
{
  return (d1>0)?1:-1;
}

/*! This function returns the maximum of two given doubles.
  \param d1 first parameter
  \param d2 second parameter
  \return the maximum of these two parameters */
double max( double d1, double d2 )
{
  return (d1>d2)?d1:d2;
}

/*! This function returns the minimum of two given doubles.
  \param d1 first parameter
  \param d2 second parameter
  \return the minimum of these two parameters */
double min( double d1, double d2 )
{
  return (d1<d2)?d1:d2;
}


/**
 * Mohan::Function that finds the minmax of three values, x, min limit
 * and max limit...
 */
double minmax( double minx, double x, double maxx ) {
  return( min( max( minx, x ), maxx ) );
}


/*! This function returns the rounded value of a give double.  
  \param d is the input number which is double
  \returns the rounded value of d which is an int */
int Round( double d ) {
  if( !finite(d) )
    return (int)(-1.0E10);
  else if( d > 0 )
    return (int)( d + 0.5 );
  else
    return (int)( d - 0.5 );
}



/*! This function converts an angle in radians to the corresponding angle in
  degrees.
  \param ang an angle in radians
  \return the corresponding angle in degrees */
AngDeg Rad2Deg( AngRad ang )
{
  return ( ang * 180 / M_PI );
}

/*! This function converts an angle in degrees to the corresponding angle in
  radians.
  \param ang an angle in degrees
  \return the corresponding angle in radians */
AngRad Deg2Rad( AngDeg ang )
{
  return ( ang * M_PI / 180 );
}

/*! This function returns the cosine of a given angle in degrees using the
  built-in cosine function that works with angles in radians.
  \param ang an angle in degrees
  \return the cosine of the given angle */
double cosDeg( AngDeg ang )
{
  return ( cos( Deg2Rad( ang ) ) );
}

/*! This function returns the sine of a given angle in degrees using the
  built-in sine function that works with angles in radians.
  \param ang an angle in degrees
  \return the sine of the given angle */
double sinDeg( AngDeg ang )
{
  return ( sin( Deg2Rad( ang ) ) );
}

/*! This function returns the tangent of a given angle in degrees using the
  built-in tangent function that works with angles in radians.
  \param ang an angle in degrees
  \return the tangent of the given angle */
double tanDeg( AngDeg ang )
{
  return ( tan( Deg2Rad( ang ) ) );
}

/*! This function returns the principal value of the arc tangent of d in 
  degrees using the built-in arc tangent function which returns this value 
  in radians.
  \param d a double value
  \return the arc tangent of the given value in degrees */
AngDeg atanDeg( double d )
{
  return ( Rad2Deg( atan( d ) ) );
}

/*! This function returns the principal value of the arc tangent of y/x in
  degrees using the signs of both arguments to determine the quadrant of the
  return value. For this the built-in 'atan2' function is used which returns
  this value in radians.
  \param y a double value
  \param x a double value
  \return the arc tangent of y/x in degrees taking the signs of x and y into
  account */
double atan2Deg( double y, double x )
{
  if ( fabs( y ) < EPSILON && fabs( x ) < EPSILON )
    return ( 0.0 );

  return ( Rad2Deg( atan2( y, x ) ) );
}

/*! This function returns the principal value of the arc cosine of d in degrees
  using the built-in arc cosine function which returns this value in radians.
  \param d a double value
  \return the arc cosine of the given value in degrees */
AngDeg acosDeg( double d )
{
  if ( d >= 1 )
    return ( 0.0 );
  else if ( d <= -1 )
    return ( 180.0 );

  return ( Rad2Deg( acos( d ) ) );
}

/*! This function returns the principal value of the arc sine of x in degrees
  using the built-in arc sine function which returns this value in radians.
  \param d a double value
  \return the arc sine of the given value in degrees */
AngDeg asinDeg( double d )
{
  if ( d >= 1 )
    return ( 90.0 );
  else if ( d <= -1 )
    return ( -90.0 );

  return ( Rad2Deg( asin( d ) ) );
}

/*! This function returns a boolean value which indicates whether the value
  'ang' (from interval [-180..180] lies in the interval [angMin..angMax].
  Examples: isAngInInterval( -100, 4, -150) returns false
  isAngInInterval(   45, 4, -150) returns true
  \param ang angle that should be checked
  \param angMin minimum angle in interval
  \param angMax maximum angle in interval
  \return boolean indicating whether ang lies in [angMin..angMax] */
bool isAngInInterval( AngDeg ang, AngDeg angMin, AngDeg angMax )
{
  // convert all angles to interval 0..360
  if ( ( ang    + 360 ) < 360 ) ang    += 360;
  if ( ( angMin + 360 ) < 360 ) angMin += 360;
  if ( ( angMax + 360 ) < 360 ) angMax += 360;

  if ( angMin < angMax ) 
    return angMin < ang && ang < angMax ;
  else                  
    return !( angMax < ang && ang < angMin );
}

/*! This method returns the bisector (average) of two angles. It deals
  with the boundary problem, thus when 'angMin' equals 170 and 'angMax'
  equals -100, -145 is returned.
  \param angMin minimum angle [-180,180]
  \param angMax maximum angle [-180,180]
  \return average of angMin and angMax. */
AngDeg getBisectorTwoAngles( AngDeg angMin, AngDeg angMax )
{
  // separate sine and cosine part to circumvent boundary problem
  return normalizeAngle(
			atan2Deg( (sinDeg( angMin) + sinDeg( angMax ) )/2.0,
				  (cosDeg( angMin) + cosDeg( angMax ) )/2.0 ) );
}


/*****************/
/* Class Point2D */
/*****************/
Point2D::Point2D( double x, double y, CoordSystemT cs ) {
  setPoint( x, y, cs );
}

/*! Overloaded version of unary minus operator for Point2Ds. It returns the
  negative Point2D, i.e. both the x- and y-coordinates are multiplied by
  -1. The current Point2D itself is left unchanged.
  \return a negated version of the current Point2D */
Point2D Point2D::operator - () const
{
  return ( Point2D( -m_x, -m_y ) );
}

/*! Overloaded version of the binary plus operator for adding a given double
  value to a Point2D. The double value is added to both the x- and
  y-coordinates of the current Point2D. The current Point2D itself is
  left unchanged.
  \param d a double value which has to be added to both the x- and
  y-coordinates of the current Point2D
  \return the result of adding the given double value to the current
  Point2D */
Point2D Point2D::operator + ( const double &d ) const
{
  return ( Point2D( m_x + d, m_y + d ) );
}

/*! Overloaded version of the binary plus operator for Point2Ds. It returns
  the sum of the current Point2D and the given Point2D by adding their
  x- and y-coordinates. The Point2Ds themselves are left unchanged.
  \param p a Point2D
  \return the sum of the current Point2D and the given Point2D */
Point2D Point2D::operator + ( const Point2D &p ) const
{
  return ( Point2D( m_x + p.m_x, m_y + p.m_y ) );
}

/*! Overloaded version of the binary minus operator for subtracting a given
  double value from a Point2D. The double value is subtracted from both
  the x- and y-coordinates of the current Point2D. The current Point2D
  itself is left unchanged.
  \param d a double value which has to be subtracted from both the x- and
  y-coordinates of the current Point2D
  \return the result of subtracting the given double value from the current
  Point2D */
Point2D Point2D::operator - ( const double &d ) const
{
  return ( Point2D( m_x - d, m_y - d ) );
}

/*! Overloaded version of the binary minus operator for Point2Ds. It returns
  the difference between the current Point2D and the given Point2D by
  subtracting their x- and y-coordinates. The Point2Ds themselves are left
  unchanged.
  \param p a Point2D
  \return the difference between the current Point2D and the given
  Point2D */
Point2D Point2D::operator - ( const Point2D &p ) const
{
  return ( Point2D( m_x - p.m_x, m_y - p.m_y ) );
}

/*! Overloaded version of the multiplication operator for multiplying a
  Point2D by a given double value. Both the x- and y-coordinates of the
  current Point2D are multiplied by this value. The current Point2D
  itself is left unchanged.
  \param d the multiplication factor
  \return the result of multiplying the current Point2D by the given
  double value */
Point2D Point2D::operator * ( const double &d ) const
{
  return ( Point2D( m_x * d, m_y * d  ) );
}

/*! Overloaded version of the multiplication operator for Point2Ds. It
  returns the product of the current Point2D and the given Point2D by
  multiplying their x- and y-coordinates. The Point2Ds themselves are left
  unchanged.
  \param p a Point2D
  \return the product of the current Point2D and the given Point2D */
Point2D Point2D::operator * ( const Point2D &p ) const
{
  return ( Point2D( m_x * p.m_x, m_y * p.m_y ) );
}

/*! Overloaded version of the division operator for dividing a Point2D by a
  given double value. Both the x- and y-coordinates of the current Point2D
  are divided by this value. The current Point2D itself is left unchanged.
  \param d the division factor
  \return the result of dividing the current Point2D by the given double
  value */
Point2D Point2D::operator / ( const double &d ) const
{
  return ( Point2D( m_x / d, m_y / d  ) );
}

/*! Overloaded version of the division operator for Point2Ds. It returns the
  quotient of the current Point2D and the given Point2D by dividing
  their x- and y-coordinates. The Point2Ds themselves are left unchanged.
  \param p a Point2D
  \return the quotient of the current Point2D and the given Point2D */
Point2D Point2D::operator / ( const Point2D &p ) const
{
  return ( Point2D( m_x / p.m_x, m_y / p.m_y ) );
}

/*! Overloaded version of the assignment operator for assigning a given double
  value to both the x- and y-coordinates of the current Point2D. This
  changes the current Point2D itself.
  \param d a double value which has to be assigned to both the x- and
  y-coordinates of the current Point2D */
void Point2D::operator = ( const double &d )
{
  m_x = d;
  m_y = d;
}

/*! Overloaded version of the sum-assignment operator for Point2Ds. It
  returns the sum of the current Point2D and the given Point2D by
  adding their x- and y-coordinates. This changes the current Point2D
  itself.
  \param p a Point2D which has to be added to the current Point2D */
void Point2D::operator +=( const Point2D &p )
{
  m_x += p.m_x;
  m_y += p.m_y;
}

/*! Overloaded version of the sum-assignment operator for adding a given double
  value to a Point2D. The double value is added to both the x- and
  y-coordinates of the current Point2D. This changes the current
  Point2D itself.
  \param d a double value which has to be added to both the x- and
  y-coordinates of the current Point2D */
void Point2D::operator += ( const double &d )
{
  m_x += d;
  m_y += d;
}

/*! Overloaded version of the difference-assignment operator for Point2Ds.
  It returns the difference between the current Point2D and the given
  Point2D by subtracting their x- and y-coordinates. This changes the
  current Point2D itself.
  \param p a Point2D which has to be subtracted from the current
  Point2D */
void Point2D::operator -=( const Point2D &p )
{
  m_x -= p.m_x;
  m_y -= p.m_y;
}

/*! Overloaded version of the difference-assignment operator for subtracting a
  given double value from a Point2D. The double value is subtracted from
  both the x- and y-coordinates of the current Point2D. This changes the
  current Point2D itself.
  \param d a double value which has to be subtracted from both the x- and
  y-coordinates of the current Point2D */
void Point2D::operator -=( const double &d )
{
  m_x -= d;
  m_y -= d;
}

/*! Overloaded version of the multiplication-assignment operator for
  Point2Ds. It returns the product of the current Point2D and the
  given Point2D by multiplying their x- and y-coordinates. This changes
  the current Point2D itself.
  \param p a Point2D by which the current Point2D has to be
  multiplied */
void Point2D::operator *=( const Point2D &p )
{
  m_x *= p.m_x;
  m_y *= p.m_y;
}

/*! Overloaded version of the multiplication-assignment operator for multiplying
  a Point2D by a given double value. Both the x- and y-coordinates of the
  current Point2D are multiplied by this value. This changes the current
  Point2D itself.
  \param d a double value by which both the x- and y-coordinates of the
  current Point2D have to be multiplied */
void Point2D::operator *=( const double &d )
{
  m_x *= d;
  m_y *= d;
}

/*! Overloaded version of the division-assignment operator for Point2Ds. It
  returns the quotient of the current Point2D and the given Point2D by
  dividing their x- and y-coordinates. This changes the current Point2D
  itself.
  \param p a Point2D by which the current Point2D has to be divided */
void Point2D::operator /=( const Point2D &p )
{
  m_x /= p.m_x;
  m_y /= p.m_y;
}

/*! Overloaded version of the division-assignment operator for dividing a
  Point2D by a given double value. Both the x- and y-coordinates of the
  current Point2D are divided by this value. This changes the current
  Point2D itself.
  \param d a double value by which both the x- and y-coordinates of the
  current Point2D have to be divided */
void Point2D::operator /=( const double &d )
{
  m_x /= d;
  m_y /= d;
}

/*! Overloaded version of the inequality operator for Point2Ds. It
  determines whether the current Point2D is unequal to the given
  Point2D by comparing their x- and y-coordinates.
  \param p a Point2D
  \return true when either the x- or y-coordinates of the given Point2D
  and the current Point2D are different; false otherwise */
bool Point2D::operator !=( const Point2D &p ) const
{
  return ( ( m_x != p.m_x ) || ( m_y != p.m_y ) );
}

/*! Overloaded version of the inequality operator for comparing a Point2D to
  a double value. It determines whether either the x- or y-coordinate of the
  current Point2D is unequal to the given double value.
  \param d a double value with which both the x- and y-coordinates of the
  current Point2D have to be compared.
  \return true when either the x- or y-coordinate of the current Point2D
  is unequal to the given double value; false otherwise */
bool Point2D::operator !=( const double &d ) const
{
  return ( ( m_x != d ) || ( m_y != d ) );
}

/*! Overloaded version of the equality operator for Point2Ds. It determines
  whether the current Point2D is equal to the given Point2D by
  comparing their x- and y-coordinates.
  \param p a Point2D
  \return true when both the x- and y-coordinates of the given Point2D and
  the current Point2D are equal; false otherwise */
bool Point2D::operator ==( const Point2D &p ) const
{
  return ( ( m_x == p.m_x ) && ( m_y == p.m_y ) );
}

/*! Overloaded version of the equality operator for comparing a Point2D to a
  double value. It determines whether both the x- and y-coordinates of the
  current Point2D are equal to the given double value.
  \param d a double value with which both the x- and y-coordinates of the
  current Point2D have to be compared.
  \return true when both the x- and y-coordinates of the current Point2D
  are equal to the given double value; false otherwise */
bool Point2D::operator ==( const double &d ) const
{
  return ( ( m_x == d ) && ( m_y == d ) );
}

/*! Overloaded version of the C++ output operator for output. 
  \param os output stream to which information should be written
  \param p a Point2D which must be printed
  \return output stream containing (x,y) */
ostream& operator << ( ostream &os, Point2D p )
{
  return ( os << "( " << p.m_x << ", " << p.m_y << " )" );
}

/*! Set method for the x-coordinate of the current Point2D.
  \param x a double value representing a new x-coordinate */
void Point2D::setX( double x )
{
  m_x = x;
}

/*! Get method for the x-coordinate of the current Point2D.
  \return the x-coordinate of the current Point2D */
double Point2D::getX() const
{
  return ( m_x );
}

/*! Set method for the y-coordinate of the current Point2D.
  \param y a double value representing a new y-coordinate */
void Point2D::setY( double y )
{
  m_y = y;
}

/*! Get method for the y-coordinate of the current Point2D.
  \return the y-coordinate of the current Point2D */
double Point2D::getY() const
{
  return ( m_y );
}

/*! This method (re)sets the coordinates of the current Point2D. The given
  coordinates can either be polar or Cartesian coordinates. This is indicated
  by the value of the third argument.
  \param x a double value indicating either a new Cartesian x-coordinate when
  cs=CARTESIAN or a new polar r-coordinate (distance) when cs=POLAR
  \param y a double value indicating either a new Cartesian y-coordinate when
  cs=CARTESIAN or a new polar phi-coordinate (angle) when cs=POLAR
  \param cs a CoordSystemT indicating whether x and y denote cartesian 
  coordinates or polar coordinates */
void Point2D::setPoint( double x, double y, CoordSystemT cs)
{
  if ( cs == CARTESIAN ) {
    m_x = x;
    m_y = y;
  }
  else
    *this = getPointFromPolar( x, y );
}

/*! This method determines the distance between the current Point2D and a
  given Point2D. This is equal to the magnitude (length) of the vector
  connecting the two positions which is the difference vector between them.
  \param p a Vecposition
  \return the distance between the current Point2D and the given
  Point2D */
double Point2D::getDistanceTo( const Point2D p ) const {
  return ( ( *this - p ).getMagnitude() );
}

/*! This method adjusts the coordinates of the current Point2D in such a way
  that the magnitude of the corresponding vector equals the double value which
  is supplied as an argument. It thus scales the vector to a given length by
  multiplying both the x- and y-coordinates by the quotient of the argument
  and the current magnitude. This changes the Point2D itself.
  \param d a double value representing a new magnitude
  \return the result of scaling the vector corresponding with the current
  Point2D to the given magnitude thus yielding a different Point2D */
Point2D Point2D::setMagnitude( double d )
{
  if ( getMagnitude() > EPSILON )
    ( *this ) *= ( d / getMagnitude() );

  return ( *this );
}

/*! This method determines the magnitude (length) of the vector corresponding
  with the current Point2D using the formula of Pythagoras.
  \return the length of the vector corresponding with the current
  Point2D */
double Point2D::getMagnitude() const
{
  return ( sqrt( m_x * m_x + m_y * m_y ) );
}

/*! This method determines the direction of the vector corresponding with the
  current Point2D (the phi-coordinate in polar representation) using the
  arc tangent function. Note that the signs of x and y have to be taken into
  account in order to determine the correct quadrant.
  \return the direction in degrees of the vector corresponding with the
  current Point2D */
AngDeg Point2D::getDirection() const
{
  return ( atan2Deg( m_y, m_x ) );
}

/*! This method determines whether the current Point2D is in front of a
  given Point2D, i.e. whether the x-coordinate of the current Point2D
  is larger than the x-coordinate of the given Point2D.
  \param p a Point2D to which the current Point2D must be compared
  \return true when the current Point2D is in front of the given
  Point2D; false otherwise */
bool Point2D::isInFrontOf( const Point2D &p ) const
{
  return ( m_y > p.getY() );
}

/*! This method determines whether the x-coordinate of the current Point2D
  is in front of (i.e. larger than) a given double value.
  \param y a double value to which the current x-coordinate must be compared
  \return true when the current x-coordinate is in front of the given value;
  false otherwise */
bool Point2D::isInFrontOf( const double &y ) const
{
  return ( m_y > y );
}

/*! This method determines whether the current Point2D is behind a given
  Point2D, i.e. whether the x-coordinate of the current Point2D is
  smaller than the x-coordinate of the given Point2D.
  \param p a Point2D to which the current Point2D must be compared
  \return true when the current Point2D is behind the given Point2D;
  false otherwise */
bool Point2D::isBehind( const Point2D &p ) const
{
  return ( m_y < p.getY() );
}

/*! This method determines whether the x-coordinate of the current Point2D
  is behind (i.e. smaller than) a given double value.
  \param y a double value to which the current x-coordinate must be compared
  \return true when the current x-coordinate is behind the given value; false
  otherwise */
bool Point2D::isBehind( const double &y ) const
{
  return ( m_y < y );
}

/*! This method determines whether the current Point2D is to the left of a
  given Point2D, i.e. whether the y-coordinate of the current Point2D
  is smaller than the y-coordinate of the given Point2D.
  \param p a Point2D to which the current Point2D must be compared
  \return true when the current Point2D is to the left of the given
  Point2D; false otherwise */
bool Point2D::isLeftOf( const Point2D &p ) const
{
  return ( m_x < p.getX() );
}

/*! This method determines whether the y-coordinate of the current Point2D
  is to the left of (i.e. smaller than) a given double value.
  \param x a double value to which the current y-coordinate must be compared
  \return true when the current y-coordinate is to the left of the given
  value; false otherwise */
bool Point2D::isLeftOf( const double &x ) const
{
  return ( m_x < x );
}

/*! This method determines whether the current Point2D is to the right of a
  given Point2D, i.e. whether the y-coordinate of the current Point2D
  is larger than the y-coordinate of the given Point2D.
  \param p a Point2D to which the current Point2D must be compared
  \return true when the current Point2D is to the right of the given
  Point2D; false otherwise */
bool Point2D::isRightOf( const Point2D &p ) const
{
  return ( m_x > p.getX() );
}

/*! This method determines whether the y-coordinate of the current Point2D
  is to the right of (i.e. larger than) a given double value.
  \param x a double value to which the current y-coordinate must be compared
  \return true when the current y-coordinate is to the right of the given
  value; false otherwise */
bool Point2D::isRightOf( const double &x ) const
{
  return ( m_x > x );
}

/*! This method determines whether the current Point2D is in between two
  given Point2Ds when looking in the x-direction, i.e. whether the current
  Point2D is in front of the first argument and behind the second.
  \param p1 a Point2D to which the current Point2D must be compared
  \param p2 a Point2D to which the current Point2D must be compared
  \return true when the current Point2D is in between the two given
  Point2Ds when looking in the x-direction; false otherwise */
bool Point2D::isBetweenX( const Point2D &p1, const Point2D &p2 ) const
{
  return ( isRightOf( p1 ) && isLeftOf( p2 ) );
}

/*! This method determines whether the x-coordinate of the current Point2D
  is in between two given double values, i.e. whether the x-coordinate of the
  current Point2D is in front of the first argument and behind the second.
  \param x1 a double value to which the current x-coordinate must be compared
  \param x2 a double value to which the current x-coordinate must be compared
  \return true when the current x-coordinate is in between the two given
  values; false otherwise */
bool Point2D::isBetweenX( const double &x1, const double &x2 ) const
{
  return ( isRightOf( x1 ) && isLeftOf( x2 ) );
}


/**
 * Mohan::Function returns true if the current Point2D lies in between
 * (in terms of the x coordinate) the Point2Ds sent in as arguments -
 * the two inputs need not be arranged in increasing order of
 * x-coordinates...
 */
bool Point2D::PointIsBetweenX( const Point2D &p1, const Point2D &p2 ) const {
  return( ( isRightOf( p1 ) && isLeftOf( p2 ) ) ||
	  ( isLeftOf( p1 ) && isRightOf( p2 ) ) );
}


/**
 * Mohan::Function returns true if the current Point2D lies in between
 * (in terms of the y coordinate) the Point2Ds sent in as arguments -
 * the two inputs need not be arranged in increasing order of
 * y-coordinates...
 */
bool Point2D::PointIsBetweenY( const Point2D &p1, const Point2D &p2 ) const {
  return( ( isInFrontOf( p1 ) && isBehind( p2 ) ) ||
	  ( isBehind( p1 ) && isInFrontOf( p2 ) ) );
}


/*! This method determines whether the current Point2D is in between two
  given Point2Ds when looking in the y-direction, i.e. whether the current
  Point2D is to the right of the first argument and to the left of the
  second.
  \param p1 a Point2D to which the current Point2D must be compared
  \param p2 a Point2D to which the current Point2D must be compared
  \return true when the current Point2D is in between the two given
  Point2Ds when looking in the y-direction; false otherwise */
bool Point2D::isBetweenY( const Point2D &p1, const Point2D &p2 ) const
{
  return ( isInFrontOf( p1 ) && isBehind( p2 ) );
}

/*! This method determines whether the y-coordinate of the current Point2D
  is in between two given double values, i.e. whether the y-coordinate of the
  current Point2D is to the right of the first argument and to the left
  of the second.
  \param y1 a double value to which the current y-coordinate must be compared
  \param y2 a double value to which the current y-coordinate must be compared
  \return true when the current y-coordinate is in between the two given
  values; false otherwise */
bool Point2D::isBetweenY( const double &y1, const double &y2 ) const
{
  return ( isInFrontOf( y1 ) && isBehind( y2 ) );
}

/*! This method normalizes a Point2D by setting the magnitude of the
  corresponding vector to 1. This thus changes the Point2D itself.
  \return the result of normalizing the current Point2D thus yielding a
  different Point2D */
Point2D Point2D::normalize()
{
  return ( setMagnitude( 1.0 ) );
}

/*! This method rotates the vector corresponding to the current Point2D over
  a given angle thereby changing the current Point2D itself. This is done
  by calculating the polar coordinates of the current Point2D and adding
  the given angle to the phi-coordinate in the polar representation. The polar
  coordinates are then converted back to Cartesian coordinates to obtain the
  desired result.
  \param ang an angle in degrees over which the vector corresponding to the
  current Point2D must be rotated
  \return the result of rotating the vector corresponding to the current
  Point2D over the given angle thus yielding a different Point2D */
Point2D Point2D::rotate( AngDeg ang )
{
  // determine the polar representation of the current Point2D
  double dMag    = this->getMagnitude();
  double dNewDir = this->getDirection() + ang;  // add rotation angle to phi
  setPoint( dMag, dNewDir, POLAR );          // convert back to Cartesian
  return ( *this );
}

/*! This method converts the coordinates of the current Point2D (which are
  represented in an global coordinate system with the origin at (0,0)) into
  relative coordinates in a different coordinate system (e.g. relative to a
  player). The new coordinate system is defined by the arguments to the
  method. The relative coordinates are now obtained by aligning the relative
  coordinate system with the global coordinate system using a translation to
  make both origins coincide followed by a rotation to align the axes.
  \param origin the origin of the relative coordinate frame
  \param ang the angle between the world frame and the relative frame
  (reasoning from the world frame)
  \return the result of converting the current global Point2D into a
  relative Point2D */
Point2D Point2D::globalToRelative( Point2D origin, AngDeg ang )
{
  // convert global coordinates into relative coordinates by aligning relative
  // frame and world frame. First perform translation to make origins of both
  // frames coincide. Then perform rotation to make axes of both frames coincide
  // (use negative angle since you rotate relative frame to world frame).
  *this -= origin;
  return ( rotate( -ang ) );
}

/*! This method converts the coordinates of the current Point2D (which are
  represented in a relative coordinate system) into global coordinates in
  the world frame (with origin at (0,0)). The relative coordinate system is
  defined by the arguments to the method. The global coordinates are now
  obtained by aligning the world frame with the relative frame using a
  rotation to align the axes followed by a translation to make both origins
  coincide.
  \param origin the origin of the relative coordinate frame
  \param ang the angle between the world frame and the relative frame
  (reasoning from the world frame)
  \return the result of converting the current relative Point2D into an
  global Point2D */
Point2D Point2D::relativeToGlobal( Point2D origin, AngDeg ang )
{
  // convert relative coordinates into global coordinates by aligning world
  // frame and relative frame. First perform rotation to make axes of both
  // frames coincide (use positive angle since you rotate world frame to
  // relative frame). Then perform translation to make origins of both frames
  // coincide.
  rotate( ang );
  *this += origin;
  return ( *this );
}

/*! This method returns a Point2D that lies somewhere on the vector between
  the current Point2D and a given Point2D. The desired position is
  specified by a given fraction of this vector (e.g. 0.5 means exactly in
  the middle of the vector). The current Point2D itself is left unchanged.
  \param p a Point2D which defines the vector to the current Point2D
  \param frac double representing the fraction of the connecting vector at
  which the desired Point2D lies.
  \return the Point2D that lies at fraction frac on the vector
  connecting p and the current Point2D */
Point2D Point2D::getPointOnLineFraction( Point2D &p,
					 double  frac )
{
  return ( ( *this ) * ( 1.0 - frac ) + ( p * frac ) );
}

/*! This method returns the angle between the 
  vectors made by this point and the two given points.
  \param p1 Point2D of first vector endpoint
  \param p2 Point2D of second vector endpoint
  \return the angle in degrees between the vectors
  (this - p1) and (this - p2) */ 
AngDeg Point2D::getAngleBetweenPoints( const Point2D &p1, 
				       const Point2D &p2 )
{
  Point2D v1 = *this - p1;
  Point2D v2 = *this - p2;
  Point2D vec = v1 * v2;
  double dotProduct = vec.getX() + vec.getY();
  return fabs( normalizeAngle( acosDeg( dotProduct / v1.getMagnitude() / v2.getMagnitude() ) ) );
}

/*! This method converts a polar representation of a Point2D into a
  Cartesian representation.
  \param mag a double representing the polar r-coordinate, i.e. the distance
  from the point to the origin
  \param ang the angle that the polar vector makes with the x-axis, i.e. the
  polar phi-coordinate
  \return the result of converting the given polar representation into a
  Cartesian representation thus yielding a Cartesian Point2D */
Point2D Point2D::getPointFromPolar( double mag, AngDeg ang )
{
  // cos(phi) = x/r <=> x = r*cos(phi); sin(phi) = y/r <=> y = r*sin(phi)
  return ( Point2D( mag * cosDeg( ang ), mag * sinDeg( ang ) ) );
}

/*! This method normalizes an angle. This means that the resulting angle lies
  between -180 and 180 degrees.
  \param ang the angle which must be normalized
  \return the result of normalizing the given angle */
//AngDeg Point2D::normalizeAngle( AngDeg ang )
AngDeg normalizeAngle( AngDeg ang )
{
  while( ang > 180.0  ) ang -= 360.0;
  while( ang < -180.0 ) ang += 360.0;

  return ( ang );
}


/******************/
/* Class Geometric */
/******************/

/*! A geometric series is one in which there is a constant ratio between each
  element and the one preceding it. This method determines the
  length of a geometric series given its first element, the sum of the
  elements in the series and the constant ratio between the elements.
  Normally: s = a + ar + ar^2 + ...  + ar^n
  Now: dSum = dFirst + dFirst*dRatio + dFirst*dRatio^2 + .. + dFist*dRatio^n
  \param dFirst first term of the series
  \param dRatio ratio with which the the first term is multiplied
  \param dSum the total sum of all the serie
  \return the length(n in above example) of the series */
double Geometric::getLengthGeomSeries( double dFirst, double dRatio, double dSum )
{
  if( dRatio < 0 )
    cerr << "(Geometric:getLengthGeomSeries): negative ratio" << endl;

  // s = a + ar + ar^2 + .. + ar^n-1 and thus sr = ar + ar^2 + .. + ar^n
  // subtract: sr - s = - a + ar^n) =>  s(1-r)/a + 1 = r^n = temp
  // log r^n / n = n log r / log r = n = length
  double temp = (dSum * ( dRatio - 1 ) / dFirst) + 1;
  if( temp <= 0 )
    return -1.0;
  return log( temp ) / log( dRatio ) ;
}

/*! A geometric series is one in which there is a constant ratio between each
  element and the one preceding it. This method determines the sum of a
  geometric series given its first element, the ratio and the number of steps
  in the series
  Normally: s = a + ar + ar^2 + ...  + ar^n
  Now: dSum = dFirst + dFirst*dRatio + ... + dFirst*dRatio^dSteps
  \param dFirst first term of the series
  \param dRatio ratio with which the the first term is multiplied
  \param dSum the number of steps to be taken into account
  \return the sum of the series */
double Geometric::getSumGeomSeries( double dFirst, double dRatio, double dLength)
{
  // s = a + ar + ar^2 + .. + ar^n-1 and thus sr = ar + ar^2 + .. + ar^n
  // subtract: s - sr = a - ar^n) =>  s = a(1-r^n)/(1-r)
  return dFirst * ( 1 - pow( dRatio, dLength ) ) / ( 1 - dRatio ) ;
}

/*! A geometric series is one in which there is a constant ratio between each
  element and the one preceding it. This method determines the sum of an
  infinite geometric series given its first element and the constant ratio
  between the elements. Note that such an infinite series will only converge
  when 0<r<1.
  Normally: s = a + ar + ar^2 + ar^3 + ....
  Now: dSum = dFirst + dFirst*dRatio + dFirst*dRatio^2...
  \param dFirst first term of the series
  \param dRatio ratio with which the the first term is multiplied
  \return the sum of the series */
double Geometric::getSumInfGeomSeries( double dFirst, double dRatio )
{
  if( dRatio > 1 )
    cerr << "(Geometric:CalcLengthGeomSeries): series does not converge" << endl;

  // s = a(1-r^n)/(1-r) with n->inf and 0<r<1 => r^n = 0
  return dFirst / ( 1 - dRatio );
}

/*! A geometric series is one in which there is a constant ratio between each
  element and the one preceding it. This method determines the first element
  of a geometric series given its element, the ratio and the number of steps
  in the series
  Normally: s = a + ar + ar^2 + ...  + ar^n
  Now: dSum = dFirst + dFirst*dRatio + ... + dFirst*dRatio^dSteps
  \param dSum sum of the series
  \param dRatio ratio with which the the first term is multiplied
  \param dSum the number of steps to be taken into account
  \return the first element (a) of a serie */
double Geometric::getFirstGeomSeries( double dSum, double dRatio, double dLength)
{
  // s = a + ar + ar^2 + .. + ar^n-1 and thus sr = ar + ar^2 + .. + ar^n
  // subtract: s - sr = a - ar^n) =>  s = a(1-r^n)/(1-r) => a = s*(1-r)/(1-r^n)
  return dSum *  ( 1 - dRatio )/( 1 - pow( dRatio, dLength ) ) ;
}

/*! A geometric series is one in which there is a constant ratio between each
  element and the one preceding it. This method determines the first element
  of an infinite geometric series given its first element and the constant
  ratio between the elements. Note that such an infinite series will only
  converge when 0<r<1.
  Normally: s = a + ar + ar^2 + ar^3 + ....
  Now: dSum = dFirst + dFirst*dRatio + dFirst*dRatio^2...
  \param dSum sum of the series
  \param dRatio ratio with which the the first term is multiplied
  \return the first term of the series */
double Geometric::getFirstInfGeomSeries( double dSum, double dRatio )
{
  if( dRatio > 1 )
    cerr << "(Geometric:getFirstInfGeomSeries):series does not converge" << endl;

  // s = a(1-r^n)/(1-r) with r->inf and 0<r<1 => r^n = 0 => a = s ( 1 - r)
  return dSum * ( 1 - dRatio );
}

/*! This method performs the abc formula (Pythagoras' Theorem) on the given
  parameters and puts the result in *s1 en *s2. It returns the number of
  found coordinates.
  \param a a parameter in abc formula
  \param b b parameter in abc formula
  \param c c parameter in abc formula
  \param *s1 first result of abc formula
  \param *s2 second result of abc formula
  \return number of found x-coordinates */
int Geometric::abcFormula(double a, double b, double c, double *s1, double *s2)
{
  double dDiscr = b*b - 4*a*c;       // discriminant is b^2 - 4*a*c
  if (fabs(dDiscr) < EPSILON )       // if discriminant = 0
  {
    *s1 = -b / (2 * a);              //  only one solution
    return 1;
  }
  else if (dDiscr < 0)               // if discriminant < 0
    return 0;                        //  no solutions
  else                               // if discriminant > 0
  {
    dDiscr = sqrt(dDiscr);           //  two solutions
    *s1 = (-b + dDiscr ) / (2 * a);
    *s2 = (-b - dDiscr ) / (2 * a);
    return 2;
  }
}


/****************/
/* Class Circle */
/****************/

/*! This is the constructor of a circle.
  \param pCenter first point that defines the center of circle
  \param radius the radius of the circle
  \return circle with p as center and radius as radius*/
Circle::Circle( Point2D pCenter, double radius )
{
  setCircle( pCenter, radius );
}

/*! This is the constructor of a circle which initializes a circle with a
  radius of zero. */
Circle::Circle()
{
  setCircle( Point2D(-1000.0,-1000.0), 0);
}

/*! Overloaded version of the C++ output operator for output. 
  \param os output stream to which information should be written
  \param c a Circle which must be printed
  \return output stream containing center and radius */
ostream& operator << ( ostream &os, Circle c )
{
  return ( os << "c: " << c.m_pCenter << ", r: " << c.m_radius );
}

/*! This method sets the values of the circle.
  \param pCenter new center of the circle
  \param radius new radius of the circle
  ( > 0 )
  \return bool indicating whether radius was set */
bool Circle::setCircle( Point2D pCenter, double radius )
{
  setCenter( pCenter );
  return setRadius( radius  );
}

/*! This method sets the radius of the circle.
  \param radius new radius of the circle ( > 0 )
  \return bool indicating whether radius was set */
bool Circle::setRadius( double radius )
{
  if ( radius > 0 )
  {
    m_radius = radius;
    return true;
  }
  else
  {
    m_radius = 0.0;
    return false;
  }
}

/*! This method returns the radius of the circle.
  \return radius of the circle */
double Circle::getRadius() const
{
  return m_radius;
}

/*! This method sets the center of the circle.
  \param pCenter new center of the circle
  \return bool indicating whether center was set */
void Circle::setCenter( Point2D pCenter )
{
  m_pCenter = pCenter;
}

/*! This method returns the center of the circle.
  \return center of the circle */
Point2D Circle::getCenter() const
{
  return m_pCenter;
}

/*! This method returns the circumference of the circle.
  \return circumference of the circle */
double Circle::getCircumference() const
{
  return 2.0*M_PI*getRadius();
}

/*! This method returns the area inside the circle.
  \return area inside the circle */
double Circle::getArea() const 
{
  return M_PI*getRadius()*getRadius();
}

/*! This method returns a boolean that indicates whether 'p' is located inside
  the circle.
  \param pCenter position of which should be checked whether it is located in the
  circle
  \return bool indicating whether p lies inside the circle */
bool Circle::isInside( Point2D pCenter ) const
{
  return m_pCenter.getDistanceTo( pCenter ) < getRadius() ;
}

/*! This method returns the two possible intersection points between two
  circles. This method returns the number of solutions that were found.
  \param c circle with which intersection should be found
  \param p1 will be filled with first solution
  \param p2 will be filled with second solution
  \return number of solutions. */
int Circle::getIntersectionPoints( Circle c, Point2D *p1, Point2D *p2) const
{
  double x0, y0, r0;
  double x1, y1, r1;

  x0 = getCenter().getX();
  y0 = getCenter().getY();
  r0 = getRadius();
  x1 = c.getCenter().getX();
  y1 = c.getCenter().getY();
  r1 = c.getRadius();

  double d, dx, dy, h, a, x, y, p2_x, p2_y;
    
  // first calculate distance between two centers circles P0 and P1.
  dx = x1 - x0;
  dy = y1 - y0;
  d = sqrt(dx*dx + dy*dy);

  // normalize differences
  dx /= d; dy /= d;

  // a is distance between p0 and point that is the intersection point P2
  // that intersects P0-P1 and the line that crosses the two intersection
  // points P3 and P4.
  // Define two triangles: P0,P2,P3 and P1,P2,P3.
  // with distances a, h, r0 and b, h, r1 with d = a + b
  // We know a^2 + h^2 = r0^2 and b^2 + h^2 = r1^2 which then gives
  // a^2 + r1^2 - b^2 = r0^2 with d = a + b ==> a^2 + r1^2 - (d-a)^2 = r0^2
  // ==> r0^2 + d^2 - r1^2 / 2*d
  a = (r0*r0 + d*d - r1*r1) / (2.0 * d);

  // h is then a^2 + h^2 = r0^2 ==> h = sqrt( r0^2 - a^2 )
  double arg = r0*r0 - a*a;
  h = (arg > 0.0) ? sqrt(arg) : 0.0;

  // First calculate P2
  p2_x = x0 + a * dx;
  p2_y = y0 + a * dy;

  // and finally the two intersection points
  x =  p2_x - h * dy;
  y =  p2_y + h * dx;
  p1->setPoint( x, y );
  x =  p2_x + h * dy;
  y =  p2_y - h * dx;
  p2->setPoint( x, y );
  
  return (arg < 0.0) ? 0 : ((arg == 0.0 ) ? 1 :  2);
}

/*! This method returns the size of the intersection area of two circles.
  \param c circle with which intersection should be determined
  \return size of the intersection area. */
double Circle::getIntersectionArea( Circle c ) const
{
  Point2D p1, p2, p3;
  double d, h, dArea;
  AngDeg ang;

  d = getCenter().getDistanceTo( c.getCenter() ); // dist between two centers
  if ( d > c.getRadius() + getRadius() )           // larger than sum radii
    return 0.0;                                   // circles do not intersect
  if ( d <= fabs(c.getRadius() - getRadius() ) ) {   // one totally in the other
    double dR = min( c.getRadius(), getRadius() );// return area smallest circle
    return M_PI*dR*dR;
  }

  int iNrSol = getIntersectionPoints( c, &p1, &p2 );
  if ( iNrSol != 2 )
    return 0.0;

  // the intersection area of two circles can be divided into two segments:
  // left and right of the line between the two intersection points p1 and p2.
  // The outside area of each segment can be calculated by taking the part
  // of the circle pie excluding the triangle from the center to the
  // two intersection points.
  // The pie equals pi*r^2 * rad(2*ang) / 2*pi = 0.5*rad(2*ang)*r^2 with ang
  // the angle between the center c of the circle and one of the two
  // intersection points. Thus the angle between c and p1 and c and p3 where
  // p3 is the point that lies halfway between p1 and p2.
  // This can be calculated using ang = asin( d / r ) with d the distance
  // between p1 and p3 and r the radius of the circle.
  // The area of the triangle is 2*0.5*h*d.

  p3 = p1.getPointOnLineFraction( p2, 0.5 );
  d = p1.getDistanceTo( p3 );
  h = p3.getDistanceTo( getCenter() );
  ang = asin( d / getRadius() );

  dArea = ang*getRadius()*getRadius();
  dArea = dArea - d*h;

  // and now for the other segment the same story
  h = p3.getDistanceTo( c.getCenter() );
  ang = asin( d / c.getRadius() );
  dArea = dArea + ang*c.getRadius()*c.getRadius();
  dArea = dArea - d*h;

  return dArea;
}


/****************/
/* Class Line2D */
/****************/
Line2D::Line2D() {
  m_a = 0;
  m_b = 0;
  m_c = 0;
}

/*! This constructor creates a line by given the three coefficents of the line.
  A line is specified by the formula ay + bx + c = 0.
  \param a a coefficients of the line
  \param b b coefficients of the line
  \param c c coefficients of the line */
Line2D::Line2D( double a, double b, double c )
{
  m_a = a;
  m_b = b;
  m_c = c;
}

/*! Overloaded version of the C++ output operator for output. 
  \param os output stream to which information should be written
  \param l a Line2D which must be printed
  \return output stream containing line equation */
ostream& operator << ( ostream &os, Line2D l )
{
  return ( os << l.m_a << "y + " << l.m_b << "x + " << l.m_c << " = 0" );
}

/*! This method returns the intersection point between the current Line and
  the specified line.
  \param line line with which the intersection should be calculated.
  \return Point2D position that is the intersection point. */
Point2D Line2D::getIntersection( Line2D line ) const
{
  Point2D p;
  double x, y;

  if ( getSlope() == line.getSlope() ) // lines are parallel, no intersection
  {
    return p;
  }
  if ( m_a == 0 )               // bx + c = 0 and a2*y + b2*x + c2 = 0 ==> x = -c/b
  {                          // calculate x using the current line
    x = -m_c/m_b;                // and calculate the y using the second line
    y = line.getYGivenX(x);
  }
  else if ( line.getACoefficient() == 0 )
  {                         // ay + bx + c = 0 and b2*x + c2 = 0 ==> x = -c2/b2
    x = -line.getCCoefficient()/line.getBCoefficient(); // calculate x using
    y = getYGivenX(x);       // 2nd line and calculate y using current line
  }
  // ay + bx + c = 0 and a2y + b2*x + c2 = 0
  // y = (-b2/a2)x - c2/a2
  // bx = -a*y - c =>  bx = -a*(-b2/a2)x -a*(-c2/a2) - c ==>
  // ==> a2*bx = a*b2*x + a*c2 - a2*c ==> x = (a*c2 - a2*c)/(a2*b - a*b2)
  // calculate x using the above formula and the y using the current line
  else
  {
    x = (m_a*line.getCCoefficient() - line.getACoefficient()*m_c)/
                    (line.getACoefficient()*m_b - m_a*line.getBCoefficient());
    y = getYGivenX(x);
  }

  return Point2D( x, y );
}


/*! This method calculates the intersection points between the current line
  and the circle specified with as center 'pCenter' and radius 'dRadius'.
  The number of solutions are returned and the corresponding points are put
  in the third and fourth argument of the method
  \param c circle with which intersection points should be found
  \param pSolution1 first intersection (if any)
  \param pSolution2 second intersection (if any) */
int Line2D::getCircleIntersectionPoints( Circle circle,
				       Point2D *pSolution1, Point2D *pSolution2 ) const
{
  int    iSol;
  double dSol1, dSol2;
  double h = circle.getCenter().getX();
  double k = circle.getCenter().getY();

  // line:   x = -c/b (if a = 0)
  // circle: (x-h)^2 + (y-k)^2 = r^2, with h = center.x and k = center.y
  // fill in:(-c/b-h)^2 + y^2 -2ky + k^2 - r^2 = 0
  //         y^2 -2ky + (-c/b-h)^2 + k^2 - r^2 = 0
  // and determine solutions for y using abc-formula
  if ( fabs(m_a) < EPSILON )
  {
    iSol = Geometric::abcFormula( 1, -2*k, ((-m_c/m_b) - h)*((-m_c/m_b) - h)
              + k*k - circle.getRadius()*circle.getRadius(), &dSol1, &dSol2);
    pSolution1->setPoint( (-m_c/m_b), dSol1 );
    pSolution2->setPoint( (-m_c/m_b), dSol2 );
    return iSol;
  }

  // ay + bx + c = 0 => y = -b/a x - c/a, with da = -b/a and db = -c/a
  // circle: (x-h)^2 + (y-k)^2 = r^2, with h = center.x and k = center.y
  // fill in:x^2 -2hx + h^2 + (da*x-db)^2 -2k(da*x-db) + k^2 - r^2 = 0
  //         x^2 -2hx + h^2 + da^2*x^2 + 2da*db*x + db^2 -2k*da*x -2k*db
  //                                                         + k^2 - r^2 = 0
  //         (1+da^2)*x^2 + 2(da*db-h-k*da)*x + h2 + db^2  -2k*db + k^2 - r^2 = 0
  // and determine solutions for x using abc-formula
  // fill in x in original line equation to get y coordinate
  double da = -m_b/m_a;
  double db = -m_c/m_a;

  double a = 1 + da*da;
  double b = 2*( da*db - h - k*da );
  double c = h*h + db*db - 2*k*db + k*k - circle.getRadius()*circle.getRadius();

  iSol = Geometric::abcFormula( a, b, c, &dSol1, &dSol2 );

  pSolution1->setPoint( dSol1, da*dSol1 + db );
  pSolution2->setPoint( dSol2, da*dSol2 + db );
  return iSol;
}

/*! This method returns the orthogonal line to a Point2D. This is the line
  between the specified position and the closest point on the line to this
  position.
  \param p Point2D point with which tangent line is calculated.
  \return Line line tangent to this pition */
Line2D Line2D::getOrthogonalLine( Point2D p ) const
{
  // ay + bx + c = 0 -> y = (-b/a)x + (-c/a)
  // tangent: y = (a/b)*x + C1 -> by - ax + C2 = 0 => C2 = ax - by
  // with p.y = y, p.x = x
  return Line2D( m_b, -m_a, m_a*p.getX() - m_b*p.getY() );
}


/**
 * Mohan::This method returns the absolute value of the acute angle
 * between the current line and the line sent in as input. \param
 * line: line with which the angle is to be determined. \return
 * Angdeg: angle between the current line and input line...
 */
AngDeg Line2D::getMyAngleWith( Line2D line ) const {
  double ang = 0;
  
  double m1, m2;
  
  // In our case we shall never have a case with a line that is
  // parallel to the y-axis but I am checking for that as a
  // matter-of-fact...
  // Both lines parallel to y-axis...
  if( m_a <= EPSILON && line.getACoefficient() <= EPSILON ) {
    return ang;
  }
  // Current line parallel to y-axis...
  else if( fabs(m_a) <= EPSILON ) {
    ang = atanDeg( fabs( line.getACoefficient() / line.getBCoefficient() ) );
    return ang;
  }
  // Input line parallel to y-axis...
  else if( fabs( line.getACoefficient() ) <= EPSILON ) {
    ang = atanDeg( fabs( m_a / m_b ) );
    return ang;
  }
  // Neither line parallel to the y-axis...
  else {
    m1 = - m_b / m_a;
    m2 = - line.getBCoefficient() / line.getACoefficient();
    ang = atanDeg( fabs( m2 - m1 ) / ( 1 + m1 * m2 ) );
    return fabs(ang);
  }
}




/**
 * Mohan::This method returns the intersection point between the
 * current Line and the specified line.  \param line line with which
 * the intersection should be calculated.  \return Point2D position
 * that is the intersection point. */
Point2D Line2D::getMyIntersection( Line2D line ) const {
  Point2D p(0,0);
  double x, y;

  double SLOPE_DIFF = EPSILON * 10;

  double y_num = m_b*line.getCCoefficient() - line.getBCoefficient()*m_c;
  double denom = m_a*line.getBCoefficient() - line.getACoefficient()*m_b;

  // If the two lines are parallel, return the zero vector as the
  // intersection point...
  if( fabs(denom) <= SLOPE_DIFF ) {
    return p;
  }

  double x_num = line.getACoefficient()*m_c - m_a*line.getCCoefficient();
  
  x = x_num/denom;
  y = y_num/denom;
  return Point2D( x, y );
}


/*! This method returns the closest point on a line to a given position.
  \param p point to which closest point should be determined
  \return Point2D closest point on line to 'p'. */
Point2D Line2D::getPointOnLineClosestTo( Point2D p ) const
{
  Line2D l2 = getOrthogonalLine( p );  // get tangent line
  return getMyIntersection( l2 );     // and intersection between the two lines
}

/*! This method returns the distance between a specified position and the
  closest point on the given line.
  \param p position to which distance should be calculated
  \return double indicating the distance to the line. */
double Line2D::getDistanceToPoint( Point2D p ) const
{
  return p.getDistanceTo( getPointOnLineClosestTo( p ) );
}

/*! This method determines whether the projection of a point on the current line
  lies between two other points ('point1' and 'point2') that lie on the same
  line.
  \param p point of which projection is checked.
  \param p1 first point on line
  \param p2 second point on line
  \return true when projection of 'p' lies between 'point1' and 'point2'.*/
bool Line2D::isInBetween( Point2D p, Point2D p1, Point2D p2) const
{
  p = getPointOnLineClosestTo( p ); // get closest point
  double dDist = p1.getDistanceTo( p2 ); // get distance between 2 p

  // if the distance from both points to the projection is smaller than this
  // dist, the p lies in between.
  return p.getDistanceTo( p1 ) <= dDist &&
         p.getDistanceTo( p2 ) <= dDist;
}

/*! This method calculates the y coordinate given the x coordinate
  \param x coordinate
  \return y coordinate on this line */
double Line2D::getYGivenX( double x ) const
{
 if ( m_a == 0 )
 {
   cerr << "(Line2D::getYGivenX) Cannot calculate Y coordinate: " ;
   cerr << endl;
   return 0;
 }
  // ay + bx + c = 0 ==> ay = -(b*x + c)/a
  return -(m_b*x+m_c)/m_a;
}

/*! This method calculates the x coordinate given the x coordinate
  \param y coordinate
  \return x coordinate on this line */
double Line2D::getXGivenY( double y ) const
{
 if ( m_b == 0 )
 {
   cerr << "(Line2D::getXGivenY) Cannot calculate X coordinate\n" ;
   return 0;
 }
  // ay + bx + c = 0 ==> bx = -(a*y + c)/a
  return -(m_a*y+m_c)/m_b;
}

/*! This method creates a line given two points.
  \param p1 first point
  \param p2 second point
  \return line that passes through the two specified points. */
Line2D Line2D::makeLineFromTwoPoints( Point2D p1, Point2D p2 )
{
  // 1*y + bx + c = 0 => y = -bx - c
  // with -b the direction coefficient (or slope)
  // and c = - y - bx
  double a=1.0, b, c;
  double dTemp = p2.getX() - p1.getX(); // determine the slope
  if ( fabs(dTemp) < EPSILON )
  {
    // ay + bx + c = 0 with vertical slope=> a = 0, b = 1
    a = 0.0;
    b = 1.0;
  }
  else
  {
    // y = (-b)x -c with -b the slope of the line
    a = 1.0;
    b = -(p2.getY() - p1.getY())/dTemp;
  }
  // ay + bx + c = 0 ==> c = -a*y - b*x
  c =  - a*p2.getY()  - b * p2.getX();
  return Line2D( a, b, c );
}

/*! This method creates a line given a position and an angle.
  \param p position through which the line passes
  \param ang direction
 of the line.
  \return line that goes through position 'p' with angle 'ang'. */
Line2D Line2D::makeLineFromPositionAndAngle( Point2D p, AngDeg ang )
{
  // calculate point somewhat further in direction 'angle' and make
  // line from these two points.
  return makeLineFromTwoPoints( p, p+Point2D(1,ang,POLAR));
}

/*! This method returns the a coefficient from the line ay + bx + c = 0.
  \return a coefficient of the line. */
double Line2D::getACoefficient() const
{
  return m_a;
}

/*! This method returns the b coefficient from the line ay + bx + c = 0.
  \return b coefficient of the line. */
double Line2D::getBCoefficient() const
{
 return m_b;
}

/*! This method returns the c coefficient from the line ay + bx + c = 0.
  \return c coefficient of the line. */
double Line2D::getCCoefficient() const
{
 return m_c;
}

/*! This method returns the slope of the line.  \return m the slope of
  the line. */
double Line2D::getSlope() const
{
  if ( m_a == 0 ) {
    cerr << "(Line2D:getSlope): divide by zero\n";
    return 1000;
  }
  return m_b / m_a;
}

/*! This method returns the y-intercept of the line.
  \return b the y-intercept of the line. */
double Line2D::getYIntercept() const
{
  if ( m_a == 0 ) {
    cerr << "(Line2D:getYIntercept): divide by zero\n";
    return 1000;
  }
  return -m_c / m_a;
}

/*! This method returns the x-intercept of the line.
  \return the x-intercept of the line. */
double Line2D::getXIntercept() const
{
  if ( m_b == 0 ) {
    cerr << "(Line2D:getXIntercept): divide by zero\n";
    return 0;
  }
  return -m_c / m_b;
}


/*******************/
/* Class Rectangle */
/*******************/

/*! This is the constructor of a Rectangle. Two points will be given. The
    order does not matter as long as two opposite points are given (left
    top and right bottom or right top and left bottom).
    \param p1 first point that defines corner of rectangle
    \param p2 second point that defines other corner of rectangle
    \return rectangle with 'p1' and 'p2' as opposite corners. */
Rectangle::Rectangle( Point2D p1, Point2D p2 )
{
  setRectanglePoints( p1, p2 );
}

/*! Overloaded version of the C++ output operator for output. 
  \param os output stream to which information should be written
  \param r a Rectangle which must be printed
  \return output stream containing rectangle endpoints */
ostream& operator << ( ostream &os, Rectangle r )
{
  return ( os << "[" << r.m_pTopLeft << ", " << r.m_pBottomRight << "]" );
}

/*! This method sets the upper left and right bottom point of the current
  rectangle.
  \param p1 first point that defines corner of rectangle
  \param p2 second point that defines other corner of rectangle */
void Rectangle::setRectanglePoints( Point2D p1, Point2D p2 )
{
  m_pTopLeft.setX    ( min( p1.getX(), p2.getX() ) );
  m_pTopLeft.setY    ( max( p1.getY(), p2.getY() ) );
  m_pBottomRight.setX( max( p1.getX(), p2.getX() ) );
  m_pBottomRight.setY( min( p1.getY(), p2.getY() ) );
}

/*! This method determines whether the given position lies inside the current
  rectangle.
  \param p position which is checked whether it lies in rectangle
  \return true when 'p' lies in the rectangle, false otherwise */
bool Rectangle::isInside( Point2D p ) const
{
  return p.isBetweenX( m_pTopLeft.getX(), m_pBottomRight.getX() ) &&
         p.isBetweenY( m_pBottomRight.getY(), m_pTopLeft.getY() );

}

/*! This method sets the top left position of the rectangle
  \param p new top left position of the rectangle
  \return true when update was successful */
void Rectangle::setPosTopLeft( Point2D p )
{
  m_pTopLeft = p;
}

/*! This method returns the top left position of the rectangle
  \return top left position of the rectangle */
Point2D Rectangle::getPosTopLeft() const
{
  return m_pTopLeft;
}

/*! This method sets the right bottom position of the rectangle
  \param p new right bottom position of the rectangle
  \return true when update was succesfull */
void Rectangle::setPosBottomRight( Point2D p )
{
  m_pBottomRight = p;
}

/*! This method returns the right bottom position of the rectangle
  \return top right bottom of the rectangle */
Point2D Rectangle::getPosBottomRight() const
{
  return m_pBottomRight;
}

/*! This method returns the left bottom position of the rectangle
  \return top left bottom of the rectangle */
Point2D Rectangle::getPosBottomLeft() const
{
  return Point2D( m_pTopLeft.getX(), m_pBottomRight.getY() );
}

/*! This method returns the right top position of the rectangle
  \return top right top of the rectangle */
Point2D Rectangle::getPosTopRight() const
{
  return Point2D( m_pBottomRight.getX(), m_pTopLeft.getY() );
}

Point2D Rectangle::getPosCenter() const
{
  return ( (m_pBottomRight + m_pTopLeft) / 2 );
}

Point2D Rectangle::getPosOutside() const
{
  return getPosTopRight() + 1;
}

double Rectangle::getWidth() const
{
  return ( m_pBottomRight - m_pTopLeft ).getX();
}

double Rectangle::getLength() const
{
  return ( m_pTopLeft - m_pBottomRight ).getY();
}

double Rectangle::getLeft() const
{
  return m_pTopLeft.getX();
}

double Rectangle::getRight() const
{
  return m_pBottomRight.getX();
}

double Rectangle::getTop() const
{
  return m_pTopLeft.getY();
}

double Rectangle::getBottom() const
{
  return m_pBottomRight.getY();
}



/**
 * Mohan::default constructor creates 4x4 matrices of zeros - it suits
 * my purposes of generating homogeneous transformation matrices...
 */
Matrix::Matrix() {
  rows = 4;
  cols = 4;
  
  M = new double*[rows];
  assert( M != NULL );
  for ( int i = 0; i < rows; i++ ) {
    M[ i ] = new double[ cols ];
    assert( M[ i ] != NULL );
  }

  for( int i = 0; i < rows; i ++ ) {
    for( int j = 0; j < cols; j ++ ){
      M[i][j] = 0.0;
    }
  }
  M[3][3] = 1.0;
}


/**
 * Mohan::constructor creates (r x c) matrices of zeros - it can be
 * used for generating rotation/perspective view matrices...
 */
Matrix::Matrix(int r, int c) {
  rows = r;
  cols = c;
  
  M = new double*[rows];
  assert( M != NULL );
  for( int i = 0; i < rows; i ++ ) {
    M[ i ] = new double[ cols ];
    assert( M[ i ] != NULL );
  }
  
  for( int i = 0; i < rows; i ++ ) {
    for( int j = 0; j < cols; j ++ ){
      M[i][j] = 0.0;
    }
  }
  M[rows-1][cols-1] = 1.0;
}


/**
 * Copy constructor - at least I think so...
 */
Matrix::Matrix(const Matrix &Mat) {
  rows = Mat.rows;
  cols = Mat.cols;

  M = new double*[rows];
  for( int i = 0; i < rows; i ++ ) {
    M[ i ] = new double[ cols ];
  }

  for( int i = 0; i < rows; i ++ ) {
    for( int j = 0; j < cols; j ++ ){
      M[i][j] = Mat.M[i][j];
    }
  }
}

/**
 * Mohan::destructor - the first thing I wrote after the constructor
 * to ensure that I do not start leaking memory now:-)
 */
Matrix::~Matrix() {
  for ( int j = 0; j < rows; j++ )
    delete [] M[ j ];
  delete [] M;
}

/**
 * Mohan::overloaded operator== for matrices...
 */
bool Matrix::operator==( Matrix Mat2 ) {
  if( rows != Mat2.rows || cols != Mat2.cols ) 
    return false;

  for( int i = 0; i < rows; i ++ ) {
    for( int j = 0; j < cols; j ++ ) {
      if( M[i][j] != Mat2.M[i][j] )
	return false;
    }
  }
  return true;
}


/**
 * Mohan::overloaded operator= for matrices...
 */
Matrix Matrix::operator=( Matrix Mat2 ) {
  Matrix temp( rows, cols );
  
  if( rows != Mat2.rows || cols != Mat2.cols ) {
    cout << "\n Thou committeth matrix error - thy code shall crash \n";
    return temp;
  }
  
  for( int i = 0; i < rows; i ++ ) {
    for( int j = 0; j < cols; j ++ ) {
      M[i][j] = Mat2.M[i][j];
    }
  }
  return *this;
}



/**
 * Mohan::overloaded operator+ for matrices...
 */
Matrix Matrix::operator+( Matrix Mat2 ) {
  Matrix temp1( rows, cols );
  
  if( rows != Mat2.rows || cols != Mat2.cols ) {
    cout << "\n Thou committeth matrix error - thy code shall crash \n";
    return temp1;
  }
  
  Matrix temp2( Mat2.rows, Mat2.cols );
  
  for( int i = 0; i < Mat2.rows; i ++ ) {
    for( int j = 0; j < Mat2.cols; j ++ ) {
      temp2.M[i][j] = M[i][j] + Mat2.M[i][j];
    }
  }
  return temp2;
}



/**
 * Mohan::overloaded operator* for matrices...
 */
Matrix Matrix::operator*( Matrix Mat2 ) {
  Matrix temp( rows, Mat2.cols );
  
  if( cols != Mat2.rows ) {
    cerr<< "\n Thou committeth matrix error - thy code shall crash \n";
    return temp;
  }

  double sum;

  for( int i = 0; i < rows; i ++ ) {
    for( int j = 0; j < Mat2.cols; j ++ ) {
      sum = 0.0;
      for( int k = 0; k < Mat2.rows; k ++ ) {
	sum += ( M[i][k] * Mat2.M[k][j] );
      }
      temp.M[i][j] = sum;
    }
  }
  return temp;
}


/**
 * Mohan::overloaded operator* for multiplying scalars with
 * matrices...
 */
Matrix Matrix::operator*( double scalar ) {
  Matrix temp( rows, cols );
  for( int i = 0; i < rows; i ++ ) {
    for( int j = 0; j < cols; j ++ ) {
      temp.M[i][j] = scalar * M[i][j];
    }
  }
  return temp;
}



/**
 * Mohan::finds transpose of matrices...
 */
Matrix Matrix::Transpose() {
  Matrix temp(cols, rows);

  if( rows == 1 && cols == 1 )
    return *this;
  
  for( int i = 0; i < cols; i ++ ) {
    for( int  j = 0; j < rows; j ++ ) {
      temp.M[i][j] = M[j][i];
    }
  }
  return temp;
}

/**
 * Mohan::returns the "rotation" part of the homogeneous transform
 * (this is a 3x3 matrix)...
 */
Matrix Matrix::GetRotationMatrix() {
  Matrix temp(3,3);

  if( cols != 4 || rows !=4 ) {
    cerr << "\n Thou committeth matrix error - thy code shall crash \n";
    return temp;
  }
  
  for( int i = 0; i < 3; i++ ) {
    for( int j = 0; j < 3; j++ ) {
      temp.M[i][j] = M[i][j];
    }
  }
  return temp;
}

/**
 * Mohan::returns the "translation" part of the homogeneous transform
 * (this is a 3x1 vector)...
 */
Matrix Matrix::GetTranslationVector() {
  Matrix temp(3,1);
  
  if( cols != 4 || rows !=4 ) {
    cerr << "\n Thou committeth matrix error - thy code shall crash \n";
    return temp;
  }
  
  for( int i = 0; i < 3; i++ ) {
    temp.M[i][0] = M[i][3];
  }
  return temp;

}

/**
 * Mohan::returns the "translation" part of the homogeneous transform
 * (this is a 4x1 homogeneous vector)...
 */
Matrix Matrix::GetTranslationMatrix() {
  Matrix temp(4,1);
  
  if( cols != 4 || rows !=4 ) {
    cerr << "\n Thou committeth matrix error - thy code shall crash \n";
    return temp;
  }
  
  for( int i = 0; i < rows; i++ ) {
    temp.M[i][0] = M[i][3];
  }
  return temp;

}


/**
 * Mohan::sets the rotation part of the homogeneous transform based on
 * the given axis of rotation and the angle of rotation (OK, IN
 * DEGREES :-( ). axis = 0 => x-axis, axis = 1 => y-axis, axis = 2 =>
 * z-axis...
 */
void Matrix::Rotate_Along_Axis( int axis, int angle ) {
  //To set the rotation matrix you should have at least 3 rows and 3
  //colums and number of rows should be equal to number of columns...
  if( cols < 3 || rows < 3 || cols != rows ) {
    cerr << "\n Thou committeth matrix error - thy code shall crash \n";
  }

  //Rotation about x-axis...
  if( axis == 0 ) {
    M[0][0] = 1;               M[0][1] = 0;                M[0][2] = 0;
    M[1][0] = 0;               M[1][1] = cosDeg(angle);    M[1][2] = -sinDeg(angle);
    M[2][0] = 0;               M[2][1] = sinDeg(angle);    M[2][2] = cosDeg(angle);
  }

  //Rotation about y-axis...
  else if( axis == 1 ) {
    M[0][0] = cosDeg(angle);   M[0][1] = 0;                M[0][2] = sinDeg(angle);
    M[1][0] = 0;               M[1][1] = 1;                M[1][2] = 0;
    M[2][0] = -sinDeg(angle);  M[2][1] = 0;                M[2][2] = cosDeg(angle);
  }

  //Rotation about z-axis...
  else if( axis == 2 ) {
    M[0][0] = cosDeg(angle);   M[0][1] = -sinDeg(angle);   M[0][2] = 0;
    M[1][0] = sinDeg(angle);   M[1][1] = cosDeg(angle);    M[1][2] = 0;
    M[2][0] = 0;               M[2][1] = 0;                M[2][2] = 1;
  }

  //You should not be using Matrices...
  else {
    cerr << "\n Invalid option - you need a course in Basic Graphics:-) \n";
  }
}


/**
 * Mohan::sets the rotation part of the homogeneous transform based on
 * the given rotation matrix input...
 */
void Matrix::SetRotation( Matrix rotate ) {
  if( rotate.rows != 3 || rotate.cols != 3 || rows != 4 || cols != 4 ) {
    cerr << "\n Thou committeth matrix error - thy code shall crash \n";
  }
  
  for( int i = 0; i < 3; i ++ ) {
    for( int j = 0; j < 3; j ++ ) {
      M[i][j] = rotate.M[i][j];
    }
  }
}


/**
 * Mohan::Sets the translation vector part of the matrix based on the
 * 3x1 / 4x1 input vector matrix...
 */
void Matrix::SetTranslation( Matrix trans ){
  if( trans.rows < 3 || trans.cols > 1 || rows != 4 || cols != 4 ) {
    cerr << "\n Thou committeth matrix error - thy code shall crash \n";
  }

  M[0][3] = trans.M[0][0];
  M[1][3] = trans.M[1][0];
  M[2][3] = trans.M[2][0];
}


/**
 * Mohan::Performs the inverse of a homogeneous transform matrix
 * (which calls this function) - THIS DOES NOT INVERT ANY OTHER FORM
 * OF MATRIX!!!!
 */
Matrix Matrix::Inverse() {
  if( rows < 3 || cols < 3 || rows != cols ) {
    cerr << "\n Thou committeth matrix error - thy code shall crash \n";
  }
  
  Matrix temp(rows, cols);

  // If it is just the rotation matrix return the transpose...
  if( rows == 3 ) {
    return ( (*this).GetRotationMatrix() ).Transpose();
  }
  
  // If it is a homogeneous transform matrix, invert it using standard
  // technique...
  temp.SetRotation( ( (*this).GetRotationMatrix() ).Transpose() );
  temp.SetTranslation( temp.GetRotationMatrix() * ( (*this).GetTranslationVector() ) * (-1.0) );
  return temp;
}


/**
 * Mohan::prints the matrix out - only for debugging...
 */
void Matrix::PrintMatrix( ) {
  cout << "\n";
  for( int i = 0; i < rows; i ++ ) {
    for( int  j = 0; j < cols; j ++ ) {
      cout << M[i][j] << "\t";
    }
    cout << "\n";
  }
}



/**
 * Mohan::Functions that perform operations on Gaussians - currently
 * used for opponents alone but can be used for any observation
 * represented as a Gaussian distribution...
 */

/**
 * Mohan::Function that merges two gaussian distributions - done in
 * the standard manner as in Kalman Filter updates...
 */
Gauss GaussianMerge( Gauss newObs, Gauss oldObs ) {
  Gauss retGauss;

  Point2D varsum = ( newObs.stdev * newObs.stdev ) + ( oldObs.stdev * oldObs.stdev );
  retGauss.mean = ( ( ( newObs.stdev * newObs.stdev ) / varsum ) * oldObs.mean + 
		    ( ( oldObs.stdev * oldObs.stdev ) / varsum ) * newObs.mean );

  Point2D retvar =  ( ( ( oldObs.stdev * oldObs.stdev ) * ( newObs.stdev * newObs.stdev ) ) 
		      / varsum );
		      
  retGauss.stdev.setPoint( sqrt( retvar.getX() ), sqrt( retvar.getY() ) );

  return retGauss;
}


/**
 * Mohan::Function that degrades a gaussian distribution (reduces our
 * certainty in the observation) - based on estimates of object
 * velocities and current positions...
 */
Gauss GaussianDegrade( Gauss knownval, Time prevT, Time currT, 
		       double x_speed, double y_speed, double theta_speed, 
		       double Variance_Degrade_Factor ) {
  Gauss retOpp;
  Point2D tempvar = ( knownval.stdev * knownval.stdev + 
		      Variance_Degrade_Factor * ( currT - prevT ) );
  retOpp.mean = knownval.mean;
  retOpp.stdev.setPoint( sqrt( fabs( tempvar.getX() ) ), 
			 sqrt( fabs( tempvar.getY() ) ) );
  return retOpp;
}


/***********/
/* Testing */
/***********/
/*
#include<iostream.h>

int main( void )
{
  double dFirst = 1.0;
  double dRatio = 2.5;
  double dSum   = 63.4375;
  double dLength = 4.0;

  printf( "sum: %f\n", Geometric::getSumGeomSeries( dFirst, dRatio, dLength));
  printf( "length: %f\n", Geometric::getLengthGeomSeries( dFirst, dRatio, dSum));
}

int main( void )
{
  Line2D l1(1,-1,3 );
  Line2D l2(1,-0.2,10 );
 Line2D l3 = Line2D::makeLineFromTwoPoints( Point2D(1,-1), Point2D(2,-2) );
 l3.show();
 cout << endl;
 l1.show();
 l2.show();
  l1.getIntersection( l2 ).show();
}


int main( void )
{
  Line2D l( 1, -1, 0 );
  Point2D s1, s2;
  int i = l.getCircleIntersectionPoints( Circle( Point2D(1,1),1) &s1, &s2 );
  printf( "number of solutions: %d\n", i );
  if ( i == 2 )
  {
    cout << s1 << " " << s2 ;
  }
  else if ( i == 1 )
  {
    cout << s1;
  }
  cout << "line: " << l;
}

int main( void )
{
  Circle c11( Point2D( 10, 0 ), 10);
  Circle c12( Point2D( 40, 3 ), 40 );
  Circle c21( Point2D(  0,0 ), 5);
  Circle c22( Point2D(  3,0 ), 40 );

  Point2D p1, p2;

  cout << c11.getIntersectionArea( c21 ) << endl;
  cout << c12.getIntersectionArea( c21 ) << endl;
  cout << c22.getIntersectionArea( c11 ) << endl;
  cout << c12.getIntersectionArea( c22 ) << endl;
  return 0;
}

int main( void )
{
  cout << getBisectorTwoAngles( -155.3, 179.0 ) << endl;
  cout << getBisectorTwoAngles( -179.3, 179.0 ) << endl;
}
*/
