#ifndef _GEOMETRIC_H
#define _GEOMETRIC_H

#include <iostream>
//#include <OPENR/OSyslog.h>
#include "math.h"
#include "Constants.h"

using namespace std;

class Point2D;
typedef Point2D Vector2D; /*!< Type definition for two-dimensional vectors. */
typedef double AngRad;    /*!< Type definition for angles in radians. */
typedef double AngDeg;    /*!< Type definition for angles in degrees. */

#define EPSILON 0.0001  /*!< Value used for floating point equality tests. */

// auxiliary numeric functions for determining the
// maximum and minimum of two given double values and the sign of a value
double max ( double d1, double d2 );
double min ( double d1, double d2 );

/**
 * Mohan::Function that finds the minmax of three values, x, min limit
 * and max limit...
 */
double minmax( double minx, double x, double maxx );

int    sign( double d1 );
int Round( double d );

// auxiliary goniometric functions which enable you to
// specify angles in degrees rather than in radians
AngDeg Rad2Deg ( AngRad ang );
AngRad Deg2Rad ( AngDeg ang );
double cosDeg  ( AngDeg ang );
double sinDeg  ( AngDeg ang );
double tanDeg  ( AngDeg ang );
AngDeg atanDeg ( double d );
double atan2Deg( double y, double x );
AngDeg acosDeg ( double d );
AngDeg asinDeg ( double d );

AngDeg normalizeAngle( AngDeg ang );

// various goniometric functions
bool   isAngInInterval     ( AngDeg ang, AngDeg angMin, AngDeg angMax );
AngDeg getBisectorTwoAngles( AngDeg angMin, AngDeg angMax );

/*! CoordSystem is an enumeration of the different specified coordinate systems.*/
enum CoordSystemT 
{
  CARTESIAN,
  POLAR
} ;


/*****************/
/* Class Point2D */
/*****************/

/*! This class contains an x- and y-coordinate of a point (x,y) as member
  data and methods which operate on this position. */
class Point2D
{
 private:
  double m_x; /*!< x-coordinate of this position */
  double m_y; /*!< y-coordinate of this position */

 public:
  // constructor for Point2D class
  Point2D( double       x = 0,
           double       y = 0,
           CoordSystemT cs = CARTESIAN);

  // overloaded arithmetic operators
  Point2D operator -  () const;
  Point2D operator +  ( const double  &d ) const;
  Point2D operator +  ( const Point2D &p ) const;
  Point2D operator -  ( const double  &d ) const; 
  Point2D operator -  ( const Point2D &p ) const;
  Point2D operator *  ( const double  &d ) const;
  Point2D operator *  ( const Point2D &p ) const;
  Point2D operator /  ( const double  &d ) const;
  Point2D operator /  ( const Point2D &p ) const;
  void    operator =  ( const double  &d );
  void    operator += ( const Point2D &p );
  void    operator += ( const double  &d );
  void    operator -= ( const Point2D &p );
  void    operator -= ( const double  &d );
  void    operator *= ( const Point2D &p );
  void    operator *= ( const double  &d );
  void    operator /= ( const Point2D &p );
  void    operator /= ( const double  &d );
  bool    operator != ( const Point2D &p ) const;
  bool    operator != ( const double  &d ) const;
  bool    operator == ( const Point2D &p ) const;
  bool    operator == ( const double  &d ) const;

  // output operator
  friend ostream& operator << ( ostream &os,
				Point2D p );

  // set- and get methods for private member variables
  void   setX( double x );
  double getX() const;
  void   setY( double y );
  double getY() const;

  // set- and get methods for derived position information
  void setPoint( double       x = 0,
		 double       y = 0,
		 CoordSystemT cs = CARTESIAN );

  double  getDistanceTo( const Point2D p ) const;
  Point2D setMagnitude ( double d );
  double  getMagnitude () const;
  AngDeg  getDirection () const;

  // comparison methods for positions
  bool isInFrontOf( const Point2D &p ) const;
  bool isInFrontOf( const double  &y ) const;
  bool isBehind   ( const Point2D &p ) const;
  bool isBehind   ( const double  &y ) const;
  bool isLeftOf   ( const Point2D &p ) const;
  bool isLeftOf   ( const double  &x ) const;
  bool isRightOf  ( const Point2D &p ) const;
  bool isRightOf  ( const double  &x ) const;
  bool isBetweenX ( const Point2D &p1,
		    const Point2D &p2 ) const;
  bool isBetweenX ( const double  &x1,
		    const double  &x2 ) const;
  
  bool PointIsBetweenX( const Point2D &p1, 
			const Point2D &p2 ) const;

  bool PointIsBetweenY( const Point2D &p1, 
			const Point2D &p2 ) const;

  bool isBetweenY ( const Point2D &p1,
		    const Point2D &p2 ) const;
  bool isBetweenY ( const double  &y1,
		    const double  &y2 ) const;

  // conversion methods for positions
  Point2D normalize       ();
  Point2D rotate          ( AngDeg ang );
  Point2D globalToRelative( Point2D origin,
                            AngDeg  ang );
  Point2D relativeToGlobal( Point2D origin,
                            AngDeg  ang );

  Point2D getPointOnLineFraction( Point2D &p,
                                  double  frac );
  AngDeg  getAngleBetweenPoints ( const Point2D &p1,
				  const Point2D &p2 );

  // static class methods
  static Point2D getPointFromPolar( double mag,
                                    AngDeg ang );
  //static AngDeg  normalizeAngle   ( AngDeg ang );

} ;


/****************/
/* Class Circle */
/****************/

/*!This class represents a circle. A circle is defined by one Point2D
  (which denotes the center) and its radius. */
class Circle
{
 private:
  Point2D m_pCenter;            /*!< Center of the circle  */
  double  m_radius;            /*!< Radius of the circle  */

 public:
  // constructors for Circle class
  Circle();
  Circle( Point2D pCenter, 
	  double  radius );

  // output operator
  friend ostream& operator << ( ostream &os,
				Circle c );

  // get and set methods
  bool setCircle( Point2D pCenter, 
		  double  radius );

  bool    setRadius       ( double radius );
  double  getRadius       () const;
  void    setCenter       ( Point2D pCenter );
  Point2D getCenter       () const;
  double  getCircumference() const;
  double  getArea         () const;
  
  // calculate intersection points and area with other circle
  bool   isInside             ( Point2D p ) const;
  int    getIntersectionPoints( Circle  c, 
				Point2D *p1, 
				Point2D *p2 ) const;
  double getIntersectionArea  ( Circle c ) const;

} ;


/****************/
/* Class Line2D */
/****************/

/*!This class contains the representation of a line. A line is defined
  by the formula ay + bx + c = 0. */
class Line2D
{
  // a line is defined by the formula: ay + bx + c = 0
  double m_a; /*!< This is the a coefficient in the line ay + bx + c = 0 */
  double m_b; /*!< This is the b coefficient in the line ay + bx + c = 0 */
  double m_c; /*!< This is the c coefficient in the line ay + bx + c = 0 */

public:
  Line2D();
  Line2D( double a, double b, double c );

  // output operator
  friend ostream& operator << ( ostream &os,
				Line2D l );

  // get intersection points with this line
  Point2D getIntersection            ( Line2D line ) const;
  AngDeg getMyAngleWith              ( Line2D line ) const;
  Point2D getMyIntersection          ( Line2D line ) const;
  int     getCircleIntersectionPoints( Circle  circle,
				       Point2D *pSolution1,
				       Point2D *pSolution2 ) const;
  Line2D  getOrthogonalLine          ( Point2D p ) const;
  Point2D getPointOnLineClosestTo    ( Point2D p ) const;
  double  getDistanceToPoint         ( Point2D p ) const;
  bool    isInBetween                ( Point2D p,
				       Point2D p1,
				       Point2D p2 ) const;

  // calculate associated variables in the line
  double getYGivenX     ( double x ) const;
  double getXGivenY     ( double y ) const;
  double getACoefficient() const;
  double getBCoefficient() const;
  double getCCoefficient() const;
  double getSlope() const;
  double getYIntercept() const;
  double getXIntercept() const;

  // static methods to make a line using an easier representation.
  static Line2D makeLineFromTwoPoints       ( Point2D p1,
					      Point2D p2 );
  static Line2D makeLineFromPositionAndAngle( Point2D p,
					      AngDeg  ang );
} ;


/*******************/
/* Class Rectangle */
/*******************/

/*!This class represents a rectangle. A rectangle is defined by two Point2Ds
   the one at the upper left corner and the one at the right bottom. */
class Rectangle
{
 private:
  Point2D m_pTopLeft;     /*!< top left position of the rectangle       */
  Point2D m_pBottomRight; /*!< bottom right position of the rectangle   */

 public:
  // constructor for Rectangle class
  Rectangle( Point2D p1, Point2D p2 );

  // checks whether point lies inside the rectangle
  bool isInside( Point2D p ) const;

  // output operator
  friend ostream& operator << ( ostream &os,
				Rectangle r );

  // standard get and set methosd
  void    setRectanglePoints( Point2D p1, Point2D p2 );
  void    setPosTopLeft     ( Point2D p );
  Point2D getPosTopLeft     () const;
  void    setPosBottomRight ( Point2D p );
  Point2D getPosBottomRight () const;

  Point2D getPosBottomLeft  () const;
  Point2D getPosTopRight    () const;
  Point2D getPosCenter      () const;

  Point2D getPosOutside     () const;
  
  double getWidth() const;
  double getLength() const;
  double getRight() const;
  double getLeft() const;
  double getTop() const;
  double getBottom() const;

} ;

/******************/
/* Class Geometric */
/******************/

/*! This class contains several static methods dealing with geometry.*/
class Geometric
{
public:
  // geometric series
  static double getLengthGeomSeries( double dFirst, double dRatio, double dSum );
  static double getSumGeomSeries   ( double dFirst, double dRatio, double dLength);
  static double getSumInfGeomSeries( double dFirst, double dRatio );
  static double getFirstGeomSeries ( double dSum, double dRatio, double dLength );
  static double getFirstInfGeomSeries( double dSum, double dRatio );

  // abc formula
  static int abcFormula( double a, double b, double c, 
			 double *s1, double *s2 );
} ;


//Mohan::Class for Matrix operations;
class Matrix {
 public:
  int rows;
  int cols;
  double** M;

  // Constructors and destructor.
  Matrix(); //default 4x4 matrix constructor...
  Matrix( int r, int c ); //matrix of dimensions r x c...
  Matrix( const Matrix &Mat ); //copy constructor...
  ~Matrix(); //destructor...

  //Functions...
  //Basic functions...
  void SetRows( int r ) { rows = r; }
  void SetCols( int c ) { cols = c; }
  int GetRows() { return rows; }
  int GetCols() { return cols; }

  //Overloaded functions...
  bool operator==( Matrix Mat2 );
  Matrix operator=( Matrix Mat2 );
  Matrix operator+( Matrix Mat2 );
  Matrix operator*( Matrix Mat2 );
  Matrix operator*( double scalar );

  //Some basic matrix operations for homogeneous transforms...
  /**
   * Mohan::finds transpose of matrices...
   */
  Matrix Transpose();

  /**
   * Mohan::returns the "rotation" part of the homogeneous transform
   * (this is a 3x3 matrix)...
   */
  Matrix GetRotationMatrix();

  /**
   * Mohan::returns the "translation" part of the homogeneous transform
   * (this is a 3x1 vector)...
   */
  Matrix GetTranslationVector();

  /**
   * Mohan::returns the "translation" part of the homogeneous transform
   * (this is a 4x1 homogeneous vector)...
   */
  Matrix GetTranslationMatrix();

  /**
   * Mohan::sets the rotation part of the homogeneous transform based on
   * the given axis of rotation and the angle of rotation (OK, IN
   * DEGREES :-( ). axis = 0 => x-axis, axis = 1 => y-axis, axis = 2 =>
   * z-axis...
   */
  void Rotate_Along_Axis( int axis, int angle );

  /**
   * Mohan::sets the rotation part of the homogeneous transform based on
   * the given rotation matrix input...
   */
  void SetRotation( Matrix rotate );

  /**
   * Mohan::Sets the translation vector part of the matrix based on
   * the 3x1 / 4x1 input vector matrix...
   */
  void SetTranslation( Matrix trans );

  /**
   * Mohan::Performs the inverse of a homogeneous transform matrix
   * (which calls this function) - THIS DOES NOT INVERT ANY OTHER FORM
   * OF MATRIX!!!!
   */
  Matrix Inverse();

  //Output debug functions...
  /**
   * Mohan:: Just prints out the matrix for debugging...
   */
  void PrintMatrix( );
};



const Rectangle FIELD_BOUNDARY( Point2D( -FIELD_WIDTH / 2, FIELD_LENGTH / 2 ),Point2D( FIELD_WIDTH / 2, -FIELD_LENGTH / 2 ) );

const Rectangle GRASS_BOUNDARY( Point2D( -GRASS_WIDTH / 2, GRASS_LENGTH / 2 ),Point2D( GRASS_WIDTH / 2, -GRASS_LENGTH / 2 ) );

//pink_blue is pink on blue.
enum ObjectT {
  OBJ_PINK_BLUE,
  OBJ_PINK_YELLOW,

  OBJ_YELLOW_GOAL_LEFT,
  OBJ_YELLOW_GOAL,
  OBJ_YELLOW_GOAL_RIGHT,

  OBJ_YELLOW_PINK,
  OBJ_BLUE_PINK,

  OBJ_BLUE_GOAL_RIGHT,
  OBJ_BLUE_GOAL,
  OBJ_BLUE_GOAL_LEFT,

  NUM_LANDMARKS,        //objects 0-10...
  
  OBJ_BALL,
  OBJ_RED_ROBOT,
  OBJ_BLUE_ROBOT,     //objects 11-13...

  OBJ_BG_LINE,
  OBJ_YG_LINE,
  // OBJ_BORDER_LINE,
  OBJ_FIELD_LINE,   // Line pixels and types:objects 14-16...

  // OBJ_BF_INTERSECT,  OBJ_BB_INTERSECT,
  OBJ_FF_INTERSECT, // Line intersection types:objects 17...

  OBJ_BLUE_BLOB,
  OBJ_YELLOW_BLOB,
  
  OBJ_INVALID
};

const double beaconZ[] = { 
  BEACON_COLOR_Z, 
  BEACON_COLOR_Z, 

  GOAL_Z, 
  GOAL_Z, 
  GOAL_Z, 

  BEACON_COLOR_Z, 
  BEACON_COLOR_Z, 

  GOAL_Z, 
  GOAL_Z, 
  GOAL_Z, 

  -1, // NUM_LANDMARKS

  BALL_DIAMETER, 
  ROBOT_BODY_Y, 
  ROBOT_BODY_Y,

  -1, // Lines
  -1,
  -1,
  -1, 

  -1, // Line intersections
  -1,
  -1, 

  -1 // OBJ_INVALID
};


const Point2D landmarkPos[] = {
  Point2D( -FIELD_WIDTH / 2 - BORDER_OFFSET, -FIELD_LENGTH / 4 ),
  Point2D( -FIELD_WIDTH / 2 - BORDER_OFFSET,  FIELD_LENGTH / 4 ),
  
  Point2D( -GOAL_WIDTH / 2,  FIELD_LENGTH / 2 ),
  Point2D(               0,  FIELD_LENGTH / 2 ),
  Point2D(  GOAL_WIDTH / 2,  FIELD_LENGTH / 2 ),

  Point2D(  FIELD_WIDTH / 2 + BORDER_OFFSET,  FIELD_LENGTH / 4 ),
  Point2D(  FIELD_WIDTH / 2 + BORDER_OFFSET, -FIELD_LENGTH / 4 ),

  Point2D(  GOAL_WIDTH / 2, -FIELD_LENGTH / 2 ),
  Point2D(               0, -FIELD_LENGTH / 2 ),
  Point2D( -GOAL_WIDTH / 2, -FIELD_LENGTH / 2 ),
} ;


/**
 * Mohan::A structure to give the total Pose (position +
 * orientation)...
 */
struct Pose {
  Point2D Position;
  AngDeg Orientation;
  double Confidence;
};


/**
 *  Mohan::A brief structure definition for the opponents and other
 *  things we may decide to Gaussian merge or Kalman filter...
 */
struct Gauss {
  Point2D mean;
  Point2D stdev;
};


/**
 *  Mohan::A brief structure definition for the opponents and other
 *  things we may decide to Gaussian merge or Kalman filter...
 */
struct MHGauss {
  Gauss Gms;
  bool updated;
  bool seen;
};



/**
 * Mohan::Function that merges two gaussian distributions - done in
 * the standard manner as in Kalman Filter updates...
 */
Gauss GaussianMerge( Gauss newOpp, Gauss oldOpp );


/**
 * Mohan::Function that degrades a gaussian distribution (reduces our
 * certainty in the observation) - based on estimates of object
 * velocities and curretn positions...
 */
Gauss GaussianDegrade( Gauss oppsource, Time prevT, Time currT, double x_speed, 
		       double y_speed, double theta_speed, 
		       double Variance_Degrade_Factor );


/**
 * Mohan::Set of constants used for localization based on line
 * intersections - ONLY FIELD INTERSECTIONS EXIST AFTER REMOVAL OF THE
 * BORDERS...
 */
/* const Point2D borderfieldintersects[] = { */
/*   Point2D( MINX+BORDER_OFFSET, 0 ), */
/*   Point2D( GOAL_BOX_X_MIN, MAXY-BORDER_OFFSET ), */
/*   Point2D( GOAL_BOX_X_MAX, MAXY-BORDER_OFFSET ), */
/*   Point2D( MAXX-BORDER_OFFSET, 0 ), */
/*   Point2D( GOAL_BOX_X_MAX, MINY+BORDER_OFFSET ), */
/*   Point2D( GOAL_BOX_X_MIN, MINY+BORDER_OFFSET ) */
/* }; */

/* const Point2D borderintersects[] = { */
/*   Point2D( MINX+BORDER_OFFSET, MINY+BORDER_OFFSET ), */
/*   Point2D( MINX+BORDER_OFFSET, MAXY-BORDER_OFFSET ), */
/*   Point2D( MAXX-BORDER_OFFSET, MAXY-BORDER_OFFSET ), */
/*   Point2D( MAXX-BORDER_OFFSET, MINY+BORDER_OFFSET ) */
/* }; */

const Point2D fieldintersects[] = {
  Point2D( GOAL_BOX_X_MIN, OWN_GOAL_BOX_Y_MAX ),
  Point2D( GOAL_BOX_X_MIN, OPP_GOAL_BOX_Y_MIN ),
  Point2D( GOAL_BOX_X_MAX, OPP_GOAL_BOX_Y_MIN ),
  Point2D( GOAL_BOX_X_MAX, OWN_GOAL_BOX_Y_MAX ),

  // Copied over from borderfieldintersects...
  Point2D( MINX+BORDER_OFFSET, 0 ),
  Point2D( GOAL_BOX_X_MIN, MAXY-BORDER_OFFSET ),
  Point2D( GOAL_BOX_X_MAX, MAXY-BORDER_OFFSET ),
  Point2D( MAXX-BORDER_OFFSET, 0 ),
  Point2D( GOAL_BOX_X_MAX, MINY+BORDER_OFFSET ),
  Point2D( GOAL_BOX_X_MIN, MINY+BORDER_OFFSET ),

  // Copied over from border intersects...
  Point2D( MINX+BORDER_OFFSET, MINY+BORDER_OFFSET ),
  Point2D( MINX+BORDER_OFFSET, MAXY-BORDER_OFFSET ),
  Point2D( MAXX-BORDER_OFFSET, MAXY-BORDER_OFFSET ),
  Point2D( MAXX-BORDER_OFFSET, MINY+BORDER_OFFSET )
};



#endif // _GEOMETRIC_H

