#include "Observations.h"

/*********************/
/* Class Observation */
/*********************/

Observation::Observation( ObjectT object,
			  double prob, double dist,
			  AngDeg heading ) 
{
  M_object = object;
  M_prob = prob;
  M_dist = dist;
  M_heading = heading;
} 


ObjectT Observation::getObject()
{
  return M_object;
}

Point2D Observation::getLocation()
{
  if( isLandmark() ) {
    return landmarkPos[ M_object ];
  } 
  else {
    Point2D p;
    p.setPoint( M_dist, M_heading, POLAR );
    return( p );
  }
}

double Observation::getDistance()
{
  return M_dist;
}

AngDeg Observation::getHeading()
{
  return M_heading;
}

double Observation::getCertainty()
{
  return M_prob;
}


// Mohan::Center of the goals should not be treated as landmarks - the
// edges of goals are more reliable...
bool Observation::isLandmark() {
  return ( M_object >=0 && 
	   M_object != OBJ_YELLOW_GOAL && 
	   M_object != OBJ_BLUE_GOAL && 
	   M_object < NUM_LANDMARKS );
}


// Mohan::When the robot is really close to the goal, its center can
// be used as a landmark - better than nothing - normally goals are
// not detected as landmarks, only goal edges are used...
bool Observation::isGoal() {
  return( M_object == OBJ_BLUE_GOAL ||
	  M_object == OBJ_YELLOW_GOAL );
}

bool Observation::isBall()
{
  return M_object == OBJ_BALL;
}

// Function to check if the input observation is in fact an
// opponent...
bool Observation::isRobot() {
  return( M_object == OBJ_RED_ROBOT ||
	  M_object == OBJ_BLUE_ROBOT );
}

void Observation::setObject( ObjectT obj )
{
  M_object = obj;
}

void Observation::setDistance( double dist )
{
  M_dist = dist;
}

void Observation::setHeading( AngDeg ang )
{
  M_heading = normalizeAngle( ang );
}

void Observation::setCertainty( double prob )
{
  M_prob = ( prob < 0 ) ? 0 : ( prob > 1 ) ? 1 : prob;
}

void Observation::degradeProbability( double factor )
{
  M_prob *= factor;
}

void Observation::adjustRelative( Vector2D dPos, AngDeg dAng )
{
    Vector2D before( M_dist, M_heading, POLAR );
    Vector2D after = before - dPos;
    setDistance( after.getMagnitude() );
    setHeading( after.getDirection() - dAng );
}


/****************************/
/* Class ObservationHistory */
/****************************/

ObservationHistory::ObservationHistory( int capacity,
					Time maxTime,
					double maxDistance,
					AngDeg maxRotation,
					double minCertainty )
{
  M_capacity = capacity;
  M_history = new HistoryRecord[ M_capacity ];

  M_maxTime = maxTime;
  M_maxDistance = maxDistance;
  M_maxRotation = maxRotation;
  M_minCertainty = minCertainty;

  reset();
}

ObservationHistory::~ObservationHistory()
{
  delete [] M_history;
}

void ObservationHistory::reset()
{
  M_numElements = 0;
  M_lastHistoryMotionUpdate = -1;
}

void ObservationHistory::removeElement( int pos )
{
  if ( pos >= 0 && pos < M_numElements ) {
    for ( int i = pos; i < M_numElements - 1; i++ )
      M_history[ i ] = M_history[ i + 1 ];
    M_numElements--;
  }
}


void ObservationHistory::addObservation( Observation observation, Time time )
{
  if ( M_numElements >= M_capacity ) {
    removeOldest();
  }
 
  if ( M_numElements > M_capacity || M_numElements < 0 ) {
    cout << "\n num elements " << M_numElements << endl;
  }
  
  M_history[ M_numElements ].observation = observation;
  M_history[ M_numElements ].timestamp = time;
  M_history[ M_numElements ].distOdometer = 0;
  M_history[ M_numElements ].angOdometer = 0;
  M_numElements++;
}


void ObservationHistory::removeOlderThan( Time time )
{
  for ( int i = 0; i < M_numElements; i++ ) {
    if ( M_history[ i ].timestamp < time ) {
      removeElement( i ); i--;
    }
  }
}


void ObservationHistory::removeOldest()
{
  Time oldestTime = 1e16;
  int oldest = 0;

  for ( int i = 0; i < M_numElements; i++ ) {
    if ( M_history[ i ].timestamp < oldestTime ) {
      oldestTime = M_history[ i ].timestamp;
      oldest = i;
    }
  }

  removeElement( oldest );
}


// Remove observations if we've moved a lot since we saw them
void ObservationHistory::removeFurtherThan( double dDist, AngDeg dAng )
{
  for ( int i = 0; i < M_numElements; i++ ) {
    double dist = M_history[ i ].distOdometer;
    AngDeg ang = M_history[ i ].angOdometer;
    if ( dist > dDist || ang > dAng ) {
      removeElement( i ); i--;
    }
  }
}


void ObservationHistory::removeWorseThan( double prob )
{
  for ( int i = 0; i < M_numElements; i++ ) {
    double certainty = M_history[ i ].observation.getCertainty();
    if ( certainty < prob ) {
      removeElement( i ); i--;
    }
  }
}


bool ObservationHistory::isEmpty()
{
  return ( M_numElements == 0 );
}

int ObservationHistory::getNumObservations()
{
  return M_numElements;
}


int ObservationHistory::getCapacity()
{
  return M_capacity;
}



Observation ObservationHistory::getMerged()
{
  if ( isEmpty() ) {
    return Observation();
  }
  
  ObjectT object = M_history[ 0 ].observation.getObject();
  double prob = 0;
  double dist = 0;
  Vector2D heading;

  //cout << "Num of elements = " << M_numElements << endl;
  for ( int i = 0; i < M_numElements; i++ ) {
    Observation obs = M_history[ i ].observation;
    double p = obs.getCertainty();
    prob += p;
    dist += p * obs.getDistance();
    heading += Vector2D( p, obs.getHeading(), POLAR );
//     if( isnan( prob ) || isnan( dist ) || isnan( obs.getHeading() ) || prob == 0 ) {
//       cout << "\n values in NaN or zero=(" << p << "," << obs.getDistance() << ","
// 	   << obs.getHeading() << ")" << endl;
//     }
  }

//   if( isnan( prob ) || isnan( dist ) || isnan( heading.getX() ) || 
//       isnan( heading.getY() ) || prob == 0 ) {
//     for ( int i = 0; i < M_numElements; i++ ) {
//       Observation obs = M_history[ i ].observation;
//       cout << "\n NaN, Obs values=(" << obs.getDistance() << ","
// 	   << obs.getHeading() << "," << obs.getCertainty() << ")" << endl;
//     }
//   }
  return Observation( object, prob / M_numElements,
		      dist / prob, heading.getDirection() );
}



void ObservationHistory::adjustObservations( Vector2D dPos, 
					     AngDeg dAng )
{
  for ( int i = 0; i < M_numElements; i++ ) {
    M_history[ i ].observation.adjustRelative( dPos, dAng );
    M_history[ i ].distOdometer += dPos.getMagnitude();
    M_history[ i ].angOdometer += fabs( dAng );
  }    
}



void ObservationHistory::updateMotion( Vector2D velocity, AngDeg rotation,
				       Time time )
{
  if ( M_lastHistoryMotionUpdate < 0 ) {
    M_lastHistoryMotionUpdate = time;
    return;
  }

  Time timeDiff = time - M_lastHistoryMotionUpdate;
  M_lastHistoryMotionUpdate = time;

  Vector2D dPos = velocity * timeDiff;
  AngDeg dAng = rotation * timeDiff;

  adjustObservations( dPos, dAng );

  // TEMP REMOVAL - ICRA EXPS...
  // removeOlderThan( time );
  removeOlderThan( time - M_maxTime );
  
  removeFurtherThan( M_maxDistance, M_maxRotation );
  removeWorseThan( M_minCertainty );
}


/**
 * Mohan::Function to remove elements from the history - used in
 * localization testing...
 */
void ObservationHistory::RemoveHistory() {
  for ( int i = 0; i < M_numElements; i++ ) {
    removeElement( i ); 
    i--;
  }
}
