#ifndef _OBSERVATIONS_H
#define _OBSERVATIONS_H

#include <iostream>
#include "../Geometric.h"
#include "../Constants.h"

// Stuff that used to be in vision

/**
 * A container class for an (x, y) point.
 */
struct Point {

  double x;  /**< The x coordinate of the point. */
  double y;  /**< The y coordinate of the point. */

  /**
   * Overloading the ostream output operator.
   * \param out The stream to send data to.
   * \param p The point to output.
   * \returns The resulting ostream.
   */
  friend std::ostream & operator<<(std::ostream& out, Point& p) {
    
    out << "The point is (" << p.x << " , " << p.y << ")" << std::endl;
    return out;
    
  }

  /**
   * Overloading the istream input operator.
   * \param in The stream to get data from.
   * \param p The point to read into.
   * \returns The resulting istream.
   */
  friend std::istream & operator>>(std::istream& in, Point& p) {
    
    in >> p.x >> p.y;
    return in;
    
  }

};

/**
 * A container class to hold the coordinates of the bounding box
 * around an object in view.  The bounding box consists of two Points
 * describing the upper left and the lower right of a rectangle that
 * encloses the object.  The class also stores which object the
 * bounding box describes.
 */
struct BoundingBox {

  ObjectID ID;   /**< Which object the BoundingBox describes. */

  Point ul;      /**< The upper left coordinate of the rectangle. */
  Point lr;      /**< The lower right coordinate of the rectangle. */
  

  /**
   * A probability estimate that this bounding box is correct.
   * Added by Mohan.
   */
  double prob;   

  /**
   * Overloading the ostream output operator.
   * \param out The stream to send data to.
   * \param box The BoundingBox to output.
   * \returns The resulting ostream.
   */
  friend std::ostream & operator<<(std::ostream& out, BoundingBox& box) {

    out << "[ For ID - " << box.ID << std::endl << "Upper Left - " << box.ul
        << "Lower Right - " << box.lr << "]" << std::endl;
    return out;

  }
  
  double height;

  /**
   * Overloading the istream input operator.
   * \param in The stream to get data from.
   * \param box The BoundingBox to read into.
   * \returns The resulting istream.
   */
  friend std::istream & operator>>(std::istream& in, BoundingBox& box) {

    in >> box.ID >> box.ul >> box.lr;
    return in;

  }

};

/**
 * A container class to hold an array of BoundingBoxes.
 * Since this array is of variable length, we also keep track of the
 * size.
 * \todo Consider making this a vector.
 */
struct VisionObjects {

  int NumberOfObjects;     /**< The length of the BoundingBox array. */
  BoundingBox *ObjectInfo; /**< The BoundingBox array. */

};



class Observation
{
 private:
  ObjectT M_object;
  double M_prob;
  double M_dist;
  AngDeg M_heading;

 public:
  Observation( ObjectT object = OBJ_INVALID,
	       double prob = 0, double dist = 0,
	       AngDeg heading = 0 );

  ObjectT getObject();
  Point2D getLocation();
  double getDistance();
  AngDeg getHeading();
  double getCertainty();

  bool isLandmark();
  bool isBall();
  bool isRobot();
  
  // Mohan::When the robot is really close to the goal, its center can
  // be used as a landmark - better than nothing - normally goals are
  // not detected as landmarks, only goal edges are used...
  bool isGoal();

  void setObject( ObjectT obj );
  void setDistance( double dist );
  void setHeading( AngDeg ang );
  void setCertainty( double prob );

  void degradeProbability( double factor );
  void adjustRelative( Vector2D dPos, AngDeg dAng );
} ;


struct HistoryRecord
{
  Observation observation;
  double distOdometer;
  AngDeg angOdometer;
  Time timestamp;
} ;


class ObservationHistory
{
 private:
  HistoryRecord *M_history;
  int M_numElements;
  int M_capacity;
  Time M_maxTime;
  double M_maxDistance;
  AngDeg M_maxRotation;
  double M_minCertainty;
  Time M_lastHistoryMotionUpdate;

  void removeElement( int pos );
  void adjustObservations( Vector2D dPos, 
			   AngDeg dAng );
  
 public:
  ObservationHistory( int capacity,
		      Time maxTime,
		      double maxDistance,
		      AngDeg maxRotation,
		      double minCertainty );

  ~ObservationHistory();

  void addObservation( Observation observation, Time time );
  void removeOlderThan( Time time );
  void removeOldest();
  void removeFurtherThan( double dDist, AngDeg dAng );
  void removeWorseThan( double prob );
  void updateMotion( Vector2D velocity, AngDeg rotation,
		     Time time );
  void reset();

  // Mohan::Temp function for localization testing...
  void RemoveHistory();

  bool isEmpty();
  int getNumObservations();
  int getCapacity();
  Observation getMerged();
} ;

#endif //_OBSERVATIONS_H




