#include "PFLocalization.h"

PFLocalization::PFLocalization( WorldState *WS, bool single, bool nr )
{
  // Particles are constructed in world state
  particles = WS->particles;

  // Save options set in main()
  useSingleEstimate = single;
  noReseed = nr;

  // Construct history
  history = new ObservationHistory*[ NUM_LANDMARKS ];
  for ( int i = 0; i < NUM_LANDMARKS; i++ ) {
    history[ i ] = new ObservationHistory( HIST_CAPACITY,
					   MAX_TIME,
					   MAX_DISTANCE,
					   MAX_ROTATION,
					   MIN_CERTAINTY );
  }
} 



PFLocalization::~PFLocalization() 
{
  delete particles;
  delete [] history;
}


/**
 * Function that performs all the worldstate updates related to
 * localization...
 */
void PFLocalization::UpdateWorldState( WorldState *WS )
{
  // Get current time from world state
  Time time = WS->getTime();

  // Get odometry info for robot during last frame
  // x and y velocity in mm/ms (i.e. m/s)
  Vector2D vel = WS->getVelocity();
  // theta (angular) velocity in deg/ms
  AngDeg angVel = WS->getAngularVelocity();

  // Perform motion update on landmark histories
  for ( int i = 0; i < NUM_LANDMARKS; i++ ) {
    history[ i ]->updateMotion( vel, angVel, time );
  }

  // Get observations from high-level vision
  Observation visionObs[ MAX_OBSERVATIONS ];
  int numObservations = WS->getObservations( visionObs );

  // Update histories using new observations 
  // seen in current frame
  for( int i = 0; i < numObservations; i++ ) {
    ObjectT obj = visionObs[ i ].getObject();
    if ( visionObs[ i ].isLandmark() ) {
      history[ obj ]->addObservation( visionObs[ i ], time );
    }
  }

  // Get merged estimates from all non-empty landmark histories
  Observation mergedObs[ NUM_LANDMARKS ];
  int numMerged = 0;
  
  for ( int i = 0; i < NUM_LANDMARKS; i++ ) {
    if ( !history[ i ]->isEmpty() ) {
      mergedObs[ numMerged ] = history[ i ]->getMerged();
      numMerged++;
      }
  }

  if ( useSingleEstimate ) {
    // If "-s" option is used
    UpdateWorldStateSingleEstimate( WS,
				    vel, angVel, time,
				    mergedObs, numMerged );
  } else {
    // Otherwise
    UpdateWorldStateWithParticles( WS,
				   vel, angVel, time,
				   visionObs, numObservations,
				   mergedObs, numMerged );
  }
}

void PFLocalization::UpdateWorldStateSingleEstimate( WorldState *WS,
						     Vector2D vel, AngDeg angVel,
						     Time updateTime,
						     Observation histLandmarks[], 
						     int numHist )
{
  // Keep track of the last time an update was done
  static Time lastMotionUpdate = -1;

  if ( lastMotionUpdate < 0 ) {
    lastMotionUpdate = updateTime;
    return;
  }

  Time timeDiff = updateTime - lastMotionUpdate;
  lastMotionUpdate = updateTime;

  // Calculate motion in last frame
  // x and y change (relative to robot) in mm
  Vector2D dPos = vel * timeDiff;
  // orientation change in degrees
  AngDeg dAng = angVel * timeDiff;

  // Get position estimates from triangulation
  // using all landmarks in the history
  Particle estimates[ MAX_RESEED ];
  int numEstimates = getReseedEstimates( histLandmarks, numHist, 
					 estimates );

  // Get current pose from WorldState
  Point2D location = WS->getLocation();
  AngDeg orientation = WS->getOrientation();


  //////////////////////////////////////////////////////////
  // Insert code here to get new location and orientation //
  // given the motion of the robot and the triangulation   //
  // values
  //////////////////////////////////////////////////////////


  // Update the WorldState with the new pose
  // Note: 0.9 was just chosen arbitrarily as the confidence.
  WS->setPose( location,
	       orientation,
	       0.9 );
}

void PFLocalization::UpdateWorldStateWithParticles( WorldState *WS,
						    Vector2D vel, AngDeg angVel,
						    Time updateTime,
						    Observation seenLandmarks[], 
						    int numSeen,
						    Observation histLandmarks[], 
						    int numHist )
{
  // Move each particle according to robot's motion
  particles->updateMotion( vel, angVel, numSeen, updateTime );

  // Get position estimates from triangulation
  // using all landmarks in the history
  Particle estimates[ MAX_RESEED ];
  int numEstimates = getReseedEstimates( histLandmarks, numHist, 
					 estimates );
  
  // Update particles with seen landmarks and reseed estimates
  particles->updateObservation( seenLandmarks, numSeen,
                                estimates, numEstimates );
  
  // Use the estimate returned by particle filtering to update world
  // state...
  Particle poseEstimate = particles->getPoseEstimate();
  WS->setPose( poseEstimate.getPosition(),
	       poseEstimate.getAngle(),
	       poseEstimate.getProbability() );
}


int PFLocalization::getReseedEstimates( Observation *obs, int numObs, 
					Particle *estimates ) {

  // If reseeding is disabled with the "-n" switch,
  // don't calculate any triangulation values
  if ( noReseed ) {
    return 0;
  }
  
  int numEstimates = 0;

  //////////////////////////////////////////
  // Insert code to fill in the estimates //
  // array with good reseed values        //
  // calculated using triangulation       //
  //////////////////////////////////////////

  return numEstimates;
}


void PFLocalization::getInfoMessage( char *buffer ) {
  particles->getInfoMessage( buffer );
}
