#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <assert.h>
#include <string.h>
#include <math.h>

#include "Particles.h"

/******************/
/* Class Particle */
/******************/

Particle::Particle() {
  M_prob = -1;
}

Particle::Particle( double prob,
		    Point2D pos,
		    AngDeg ang,
		    Rectangle boundary ) {
  setParticle( prob, pos, ang, boundary );
}

double Particle::getProbability()
{
  return M_prob;
}

Point2D Particle::getPosition()
{
  return M_pos;
}

AngDeg Particle::getAngle()
{
  return M_ang;
}

void Particle::setProbability( double prob )
{
  const double MIN_PROB = 1e-4;

  M_prob = ( prob < MIN_PROB ) ? MIN_PROB : ( prob > 1 ) ? 1 : prob;  
}


void Particle::adjustProbability( double prob, double maxIncrease,
				  double maxDecrease ) {
  double newProb;
  if ( prob > M_prob + maxIncrease )
    newProb = M_prob + maxIncrease;
  else if ( prob < M_prob - maxDecrease )
    newProb = M_prob - maxDecrease;
  else
    newProb = prob;
  
  setProbability( newProb );
}


void Particle::degradeProbability( double factor )
{
  setProbability( M_prob * factor );
}

bool Particle::setPosition( Point2D pos, Rectangle boundary )
{
  if ( boundary.isInside( pos ) ) {
    M_pos = pos;
    return true;
  }
  return false;
}

void Particle::setAngle( AngDeg ang )
{
  M_ang = normalizeAngle( ang );
}


void Particle::setParticle( double prob, Point2D pos, AngDeg ang,
			    Rectangle boundary )
{
  setProbability( prob );
  setPosition( pos, boundary );
  setAngle( ang );
}


void Particle::moveRelative( Vector2D relativeDisplacement,
			     AngDeg rotation, Rectangle boundary )
{
  Point2D newPos =
    relativeDisplacement.relativeToGlobal( M_pos, M_ang );
  setPosition( newPos, boundary );
  setAngle( M_ang + rotation );
}

void Particle::moveGlobal( Vector2D globalDisplacement,
			   AngDeg rotation, Rectangle boundary )
{
  setPosition( M_pos + globalDisplacement, boundary );
  setAngle( M_ang + rotation );
}


void Particle::randomWalk( double maxDistance, AngDeg maxRotation,
			   Rectangle boundary )
{
  Vector2D dPos( maxDistance * ( 1 - M_prob ) * frand( -1, 1 ),
		 maxDistance * ( 1 - M_prob ) * frand( -1, 1 ) );
  AngDeg dAng =  maxRotation * ( 1 - M_prob ) * frand( -1, 1 );
  
  moveGlobal( dPos, dAng, boundary );
}

void Particle::placeRandomly( Rectangle boundary )
{
  Vector2D dPos( boundary.getWidth() * frand(),
		 boundary.getLength() * frand() );
  setPosition( boundary.getPosBottomLeft() + dPos,
	       boundary );
  setAngle( frand( -180, 180 ) );
}

double Particle::getDistanceToPoint( Point2D p )
{
  return M_pos.getDistanceTo( p );
}

AngDeg Particle::getAngleToPoint( Point2D p )
{
  Vector2D relativeDisplacement =
    p.globalToRelative( M_pos, M_ang );
  return relativeDisplacement.getDirection();
}

// Construct message for communication with simulator
void Particle::getInfoString( char *buffer, char ID,
			      Point2D pos, AngDeg ang )
{
  Point2D temp;
  
  switch( ID ) {
  case 'p':
    sprintf( buffer, "(p %d %d %d %d)", (int) M_pos.getX(),
	     (int) M_pos.getY(), (int) M_ang, 
	     (int) ( M_prob / PERCENT ) );
    break;
  case 'b':
    temp = M_pos;
    temp.relativeToGlobal( pos, ang );
    sprintf( buffer, "(b %d %d %d)", (int) temp.getX(),
	     (int) temp.getY(), (int) ( M_prob / PERCENT ) );
    break;
  default:
    cout << "Error: unknown ID " << ID << endl;
  }
}



/*******************/
/* Class Particles */
/*******************/

Particles::Particles() : M_boundary( Rectangle(Point2D(), Point2D()) ) 
{
  // do nothing
}

Particles::Particles( int numParticles,
		      double maxProbIncrease,
		      double maxProbDecrease,
		      double deltaDist,
		      double deltaAng,
		      double degradeFactor,
		      Rectangle boundary ) :
  M_boundary( boundary )
{
  M_numParticles = numParticles;
  M_particles = new Particle[ M_numParticles ];
  M_particlesSwap = new Particle[ M_numParticles ];

  M_maxProbIncrease = maxProbIncrease;
  M_maxProbDecrease = maxProbDecrease;

  M_deltaDist = deltaDist;
  M_deltaAng = deltaAng;
  M_degradeFactor = degradeFactor;

  reset();
}

Particles::~Particles()
{
  delete [] M_particles;
  delete [] M_particlesSwap;
}


double Particles::getProbabilitySum( Particle *p, int numParticles ) {
  double sum = 0;
  for ( int i = 0; i < numParticles; i++ ) {
    sum += p[ i ].getProbability();
  }
  return sum;
}



void Particles::updateProbabilities( Observation *observation, int numObservations )
{
  // Don't update the probabilities if there are no observations
  if ( numObservations <= 0 )
    return;

  // For each particle
  for ( int i = 0; i < M_numParticles; i++ ) {
    double newProb = 0;
    // For each observation
    for ( int j = 0; j < numObservations; j++ ) {

      // Get expected and observed angles
      AngDeg expectedAng = 
	M_particles[i].getAngleToPoint( observation[ j ].getLocation() );
      AngDeg observedAng = observation[ j ].getHeading();
      
      // Get expected and observed distances
      double expectedDist = 
	M_particles[i].getDistanceToPoint( observation[ j ].getLocation() );
      double observedDist = observation[ j ].getDistance();


      ////////////////////////////////////////////////
      // Insert code here to calculate the desired  //
      // probability of this particle, given the    //
      // expected and observed distances and angles //
      ////////////////////////////////////////////////

    }

    // Move particle's probability a little closer to 'newProb'
    M_particles[ i ].adjustProbability( newProb, M_maxProbIncrease, M_maxProbDecrease );
  }
}



// sampleParticles tries to fill numSlots slots in dst with copies of srcSize
// particles from src, whose (weight*probability)'s add up to sum.
// It may be passed a collection of particles whose weight*probability's
// don't add up to sum, in which case it takes up proportionately fewer
// slots.
int Particles::sampleParticles( Particle *src, int srcSize,
				int numSlots, double &sum,
				Particle* dst, double weight ) {
  if (sum == 0 || numSlots == 0) {
    return 0;
  }

  // mult is how many slots a particle of weight*prob 1 deserves.
  double mult = numSlots/sum; 
  
  double x = frand(0,1/mult); //mult != 0 since numSlots != 0
 
  double oldx = x;
  
  int n = 0;
  bool full = false;

  for ( int i = 0; i < srcSize; i++ ) {
    x += src[ i ].getProbability() * weight;
    
    // parts is the number of slots this particle gets.
    int parts = (int) (floor(x*mult) - floor(oldx*mult));

    for ( int j = 0; j < parts; j++ ) {
      if( n >= numSlots ) {
	full = true;
	break;
      }
      else {
	dst[ n++ ] = src[ i ];
      }
    }

    sum -= src[ i ].getProbability() * weight;
    oldx = x;
    if (full) {
      break;
    }
  }
  return n;
}


// Observation model based update
void Particles::updateObservation( Observation* observation,
				   int numObservations,
				   Particle *reseedValues,
				   int numReseedValues )
{  
  updateProbabilities( observation, numObservations );  

  double seedWeight;
  if (numReseedValues == 0) {
    seedWeight = 0;
  } 
  else {
    seedWeight = (M_numParticles * RESEED_IMPORTANCE)/numReseedValues;
  }

  double sum = getProbabilitySum( M_particles, M_numParticles ) +
    getProbabilitySum( reseedValues, numReseedValues ) * seedWeight;

  // Resample existing particles according to their weights
  int numResampled = 
    sampleParticles( M_particles, M_numParticles,
		     M_numParticles, sum, 
		     M_particlesSwap, 1 );
  
  // now add in reseed values
  // (numResampled+numReseed) should always equal NUM_ROBOT_PARTICLES
  int numReseed = sampleParticles( reseedValues, numReseedValues,
				   M_numParticles - numResampled, sum, 
				   &M_particlesSwap[ numResampled ], seedWeight ); 

  // swap particles back
  Particle *tmp = M_particles;
  M_particles = M_particlesSwap;
  M_particlesSwap = tmp;
  

  // Move particles around randomly according to their weight
  for ( int i = 0; i < M_numParticles; i++ ) {
    M_particles[ i ].randomWalk( M_deltaDist, M_deltaAng, M_boundary );
  }
}



// Motion model based update.
void Particles::updateMotion( Vector2D velocity, AngDeg rotation, 
			      int numObservations, Time updateTime )
{
  if ( M_lastMotionUpdate < 0 ) {
    M_lastMotionUpdate = updateTime;
    return;
  }

  Time timeDiff = updateTime - M_lastMotionUpdate;
  M_lastMotionUpdate = updateTime;

  Vector2D dPos = velocity * timeDiff;
  AngDeg dAng = rotation * timeDiff;

  for ( int i = 0; i < M_numParticles; i++ ) {
    M_particles[ i ].moveRelative( dPos, dAng, M_boundary );

    // If there is movement OR if no landmark is seen, degrade the
    // particle probabilities.
    if ( dPos != 0 || dAng != 0 || numObservations == 0 ) {
      M_particles[ i ].degradeProbability( M_degradeFactor );
    }
  }
}



void Particles::reset() {
  for ( int i = 0; i < M_numParticles; i++ ) {
    M_particles[ i ].setProbability( 0.1 );
    M_particles[ i ].placeRandomly( M_boundary );
  }
  M_lastMotionUpdate = -1;
}



// Mohan::Function estimates the net pose of the robot based on the
// weighted average of the particles - no longer clustering...
Particle Particles::getPoseEstimate() {
  Point2D w_possum;
  w_possum.setPoint(0, 0);
  Vector2D w_ang;

  double probval = 0;
  for( int i = 0; i < M_numParticles; i ++ ) {
    w_possum += M_particles[i].getPosition() * M_particles[i].getProbability();
    probval += M_particles[i].getProbability();
    w_ang += Vector2D( M_particles[i].getProbability(), 
		       M_particles[i].getAngle(), POLAR );
  }

  return Particle( ( probval / M_numParticles ), ( w_possum / probval ), 
		   w_ang.getDirection(), M_boundary );
}


int Particles::getNumParticles() {
  return M_numParticles;
}


// Construct message to send to simulator
void Particles::getInfoMessage( char *buffer )
{
  char pMsg[ 30 ];

  strcpy( buffer, "(info " );
  for ( int i = 0; i < M_numParticles; i++ ) {
    M_particles[ i ].getInfoString( pMsg, 'p', Point2D(), 0 );
    strcat( buffer, pMsg );
  }
  strcat( buffer, ")" );
}

