/**
 * \file WorldState.h 
 *
 * Definition of WorldState class, which holds information about the
 * state of the world
 */
#ifndef WorldState_H
#define WorldState_H

//#include "JointsAndSensors.h"
#include "Types.h"
//#include "../MovementModule/MoveParamData.h"
#include "Geometric.h"
#include "Localization/Observations.h"
//#include "Localization/BallParticles.h"
#include "Localization/Particles.h"
//#include "CommunicationInterface.h"
//#include "VisionInterface.h"
//#include "SystemTime.h"

#define TRAINING_SAMPLE_SIZE 600
#define TIME_WINDOW 1.0
#define AVG_LIMIT 40

#define COLL_FRAMES 30

/*  // custom senses (moved from JointsAndSensors 16feb04 by peggy) */
/*  enum CUSTOM_SENSE_t { */
/*    CS_BF_TAP, */
/*    CS_BM_TAP, */
/*    CS_BR_TAP, */
/*    CS_BF_PRESS, */
/*    CS_BM_PRESS, */
/*    CS_BR_PRESS, */
/*    CS_HEAD_TAP, */
/*    CS_HEAD_PRESS_LONG, */
/*    CS_CHIN_TAP, */
/*    CS_HEAD_DOUBLE_TAP, */
/*    NUM_CUSTOM_SENSES */
/*  }; */


// information about each teammate
struct TeammateState {
  PLAYER_ID_t ID;
  
  //Ball information...
  bool seesBall;
  double BallDist;
  Point2D ballPosition;
  double ballCertainty;
  int consecutiveNotBall;
  bool stopLookBall;
  // Variable that is set to make the robot go to ball...
  bool goToBall;
  // Variable that is set to imply that robot is not in a position to
  // go to the ball...
  bool cantGetToBall;

  Time lastBallMessageTime;

  //Teammate position information...
  Pose pose;
  Time lastPositionMessageTime;

  //The role of the teammate...
  role_t Role;
  // Variable that is set to true iff the robot has control over the
  // ball (chinpinchturning and/or kicking)...
  bool haveBall;

  // The state that the robot is in currently...
  game_state_t GameSt;

  // Mohan::The opponents as seen by this teammate...
  MHGauss Opps[NUM_OPPONENTS];
};




class WorldState {
 public:
  Time time;
  Time previousTime;
  Time timeInFrame;

  // DEBUG FLAG...
  Time prevBMesgTime;
  Time prevPMesgTime;

  //Ben ::
  bool isPasser;
  int tmid;
  
  // temporary for debugging -peggy
  //bool chinpinchopenloop;
  //bool seeyellowtaskside;
  //bool seeyellowvisionside;
  bool aboutToMove;

  // true if we scored, or if they scored
  bool weScored, theyScored;

  // true if we want to flash all of our LEDs like mad
  bool flashAllLeds;
  // true if we want to flash none of our LEDs 
  bool flashNoLeds;

  // Tells us if we've fallen over, and how
  FALLEN_STATE fall_direction;

  // Tells us how much battery power is remaining
  unsigned short remainingCapacity;

  // Our starting position
  Pose startingPose;

  // The robot's spread-out position...
  Pose spreadPose;

  // Whether we are the kickoff taker
  bool kickoffTaker;

  // Whether we should be doing the special set-play kick right now
  bool setPlayKickingOff;

/*    // the latest command from UTAssist */
/*    byte utassistCommand[1024]; */
/*    // the size of the latest command */
/*    int utassistCommandSize; */
/*    // the brain cycle at which we last received a UTAssist command */
/*    int utassistCommandTimestamp; */

  // general info about each teammate
  TeammateState teammateState[NUM_PLAYERS];

  // a timestamp of the last command message we received
  Time lastCommandMessage;

  // our current role
  role_t currentRole;

  // strategic go to ball
  bool strategyGoToBall;

  // Mohan::Robot's ability to return from supporting position...
  bool robotGoBack;

  // The boolean is set to true iff the robot (in behavior) stops and
  // stares at the ball because the robot and the ball are in the goal
  // box zone...
  bool stopLookBall;
  
  // The boolean set to true iff the robot for some reason (out of
  // zone, obstructed etc) cannot get to the ball though it does see
  // it...
  bool cantGetToBall;

  // An adjacency matrix
  bool alive[NUM_PLAYERS][NUM_PLAYERS];

  // The last time we've heard from our teammates.
  int lastHeard[NUM_PLAYERS];

  // information about our phsyical location on the field
  Point2D bodyPos;
  AngDeg bodyAng;
  double bodyCertainty;
  Vector2D bodyVel;
  AngDeg bodyAngVel;

  AngDeg headPan, headTilt, headTilt2;

  bool haveBall;

  //Dan: These are currently set in Movement->updateworldstate
  AngDeg bodyTiltMove; //In contrast to Mohan's Body_Tilt (below), which takes 
                       //from accelerometers, this takes from the movement command.
  double frontHeight;

  //Dan: These are set in vision
  //They are only accurate if ballInVision
  double ballLookingDirX; // Sorry; is there a Vector3D?
  double ballLookingDirY; // Btw, this is a unit vector.
  double ballLookingDirZ;
  Vector2D relBalls [BallQueueMax];
  Vector2D ballVel;
  double timeleft;  // Info from ballvel to me; computed in wsmerge
  double xcontact;
  Time   ballTimes [BallQueueMax];
  double ballDistance;
  double lastBallDistance [BallDistQueueMax];
  double ballAngle;

  //Dan: Penalty shot variable
  bool penaltyShot;

  // Mohan:: Variables that store the time when the last legal ball
  // position update was received from a teammate and the time since
  // the last such update...
  Time lastTMBallInfoTime;
  Time TSLastBallUpdate;

  double noBallInfoTime;
  double lastBallInfoTime;

  // a reference to a communication object, from which we can retrieve
  // messages
  //CommunicationInterface* communication;

  /**
   * Mohan::WorldState constructor updates the robot's ID on startup -
   * if no ID specified, throws exception...  But not anymore - Kurt
   */
  //k WorldState();
  WorldState( PLAYER_ID_t MyID );//, CommunicationInterface* com );
  
  void GetData();
  void UpdateNetworkStatus();
  void Update();
  void UpdateRelBalls(Vector2D NewBall);
  void ComputeBallVel();

  /**
   * Mohan::Function that sends the robot's information to its
   * teammates...
   */
  void UpdateToTeammates();

  /**
   * Mohan::Function that receives information from the robot's
   * teammates to be used later...
   */
  void UpdateFromTeammates();


  // Is the specified robot alive?
  bool isAlive(PLAYER_ID_t ID);
  bool isAlive(int ID);

  Time getTime();

  void setPose( Point2D pos, AngDeg ang, double certainty );
  void setVelocities( Vector2D vel, AngDeg angVel );

  void resetObservations();
  void addObservation( ObjectT obj, double dist, 
		       AngDeg ang, double certainty );

  /**
   * Mohan::A separate function to add opponent observation - similar to
   * the one above for adding other observations of landmarks etc. This
   * makes processing of opponent observations easier...
   */
  void addOppObservation( ObjectT obj, double dist,
			  AngDeg ang, double certainty );

  AngDeg getHeadPan();
  AngDeg getHeadTilt();

  Point2D getLocation();
  AngDeg getOrientation();
  double getPoseCertainty();
  Pose getMyPose();

  int getObservations( Observation *obs );

  Vector2D getVelocity();
  AngDeg getAngularVelocity();

  double distanceToPoint( Point2D p );
  AngDeg angleToPoint( Point2D p );

  void getInfoMessage( char *buffer );
  
  bool M_ID_set;
  PLAYER_ID_t M_ID;
  PLAYER_ID_t getMyID();
  int getBallID();
  int getTeamID();
  int getOpponentID();
  int getOwnGoalID();
  int getOpponentGoalID();
  int getOppGoalColor();
  
  // information from the vision module
  BoundingBox bboxes[ MAX_OBSERVATIONS ];
  int numBoundingBoxes;
  bool ballInVision;
  bool blueGoalInVision;
  bool yellowGoalInVision;
  bool pinkBlueInVision;
  bool pinkYellowInVision;
  bool yellowPinkInVision;
  bool bluePinkInVision;
  bool opponentsInVision;
  bool teammatesInVision;
  bool ownGoalInVision;
  bool opponentGoalInVision;
  BoundingBox* ballBox;
  // Observations that store the largest blob of the color of the
  // goals when the segmented image does not have enough of the bolb
  // to constitute a goal...
  Observation BGBlobObsr;
  Observation YGBlobObsr;

  BoundingBox* getBallBox();

  void UpdateBallInfo();
  int ConsecutiveNotBall;
  int newCNB;
  int ConsecutiveBall;
  bool BallKnown;
  bool SureBall;
  double getBallDistance();
  AngDeg getBallAngle();

  /**
   * Mohan::Function returns the probability of seeing ball in the
   * current frame of analysis - returns 0.0 if ball not seen...
   */
  double getBallConfidence();


  /**
   * Matrix that denotes the camera-center-to-center-of-body
   * transformation matrix as a product of the
   * camera-center-to-center-of-neck and the
   * center-of-neck-to-center-body transformation matrices...
   */
  Matrix Cambody;
  
  /**
   * Matrix denoting neck/camera rotation angles alone...
   */
  Matrix Camneck;

  /**
   * Matrix denoting robot body tilt and body roll alone...
   */
  Matrix Body;

  /**
   * Matrix denoting the constant neck to robot body center
   * transformation...
   */
  Matrix Neckbody;

  Matrix Body2Neck;

  /**
   * Body tilt and roll (DEGREES) stored in WorldState now...
   */
  double Body_Tilt, Body_Roll;

  /**
   * Save the observations in WorldState for Localization to read and
   * use...
   */
  Observation observations[ MAX_OBSERVATIONS ];
  int numObservations;


  // state of the game: our side, our kickoff, etc.
  set_play_t firstPlay;
  side_t side;
  game_state_t gameState;
  game_state_t prevGameState;
  kickoff_t kickoff;
  //int penalties[NUM_PENALTY];
  short ownScore;
  short opponentScore;
  // Mohan::Including the ball drop-in factor...
  int DropInTeam;
  int DropInTime;

  // the 2005 gamecontroller now sends player num and team num, which
  // are saved in these variables
/*    int gc_player_idx; */
/*    int gc_team_idx; */
/*    int gc_opp_team_idx; */

    // memory shared with the movement module
  //MMStatusInfo* sharedMMStatusInfo;


  // information from joints and sensors
  double joints[NUM_JOINTS];
  double sensors[NUM_SENSORS];
  // values from the last Brain cycle for the head
  double previousNeckTilt, previousHeadPan, previousHeadTilt;
  // controls how much influence new head pans will have on the head
  // pan in world state: 1.0 -> all new pan, 0.0 -> pan never changes
  double headPanLambda;
  // controls how quickly new feet colliding data will be noticed
  double ffLambda;
  // a number that represents how much the front two paw sensors have
  // been active recently.  Higher => more active, 0 => not active
  double frontFeetCount;
  // true if we've decided that our front feet are colliding with something
  bool frontFeetColliding;


  // aggregate chest sensor counter (used for close objects)
  int closeChestIRCount;
  int notCloseChestIRCount;
  bool objectCloseToChest();
  bool BallFirstFelt();
  bool BallStillFelt();

  // information for the IR ring
  unsigned char IRRing[RING_LENGTH];

  // aggregate IR sensor information
  double nearIRHistory[IR_HISTORY_SIZE];
  int nearIRHistoryIdx;
  int nearIRHighCount;
  double farIRHistory[IR_HISTORY_SIZE];
  int farIRHistoryIdx;
  int farIRHighCount;
  void addNearIRValue( double d );
  void addFarIRValue( double d );
  double nearIRAverage();
  double farIRAverage();

  //Mohan::Accelerometers being averaged...
  static const int ACC_HIST_MAX = 13;
  int AccXHistId, AccYHistId, AccZHistId;
  double AccXHist[ACC_HIST_MAX];
  double AccYHist[ACC_HIST_MAX];
  double AccZHist[ACC_HIST_MAX];
  void NewAccXVal( double x );
  void NewAccYVal( double y );
  void NewAccZVal( double z );
  double AccXAvg();
  double AccYAvg();
  double AccZAvg();

  // custom sense stuff, from JointsAndSensors
  void updateSenses();

  /**
   * Return true if the sense specified by which_sense has been
   * detected
   */
  //bool getCustomSense( CUSTOM_SENSE_t which_sense );

  // particles for ball position estimation
  //BallParticles *ballParticles;
  // particles for localization
  Particles *particles;
  // update the ball particles
  void UpdateBallParticles();
  // use particles to set some higher-level variables about ball position
  void UpdateBallEstimate();
  Rectangle* BALL_BOUNDARY;
  // holds the location of the greatest concentration of ball particles
  Point2D particleBallEstimate;
  // global version of above
  Point2D globalBallEstimate;
  // global angle to the ball
  AngDeg globalBallAngle;
  // the confidence of the particleBallEstimate
  double BallConfidence;
  // the heading in degrees of the largest clump of ball particles
  AngDeg particleBallPan;
  // half of the pan that N% of the ball particles encompass
  AngDeg particleBallPanSpread;
  

  // the time that we started the current point at
  Time pointStartTime;

  /**
   * Mohan::Function that assigns a standard deviation to the relative
   * ball position as a function of the ball distance...
   */
  Point2D GetBallStdevs( double distance );

  /**
   * Mohan::Function performs a motion update of the ball in relative
   * frame of the robot based on the movement commands that the robot
   * executes...
   */
  void Motion_Update_Gaussian();

  /**
   * Mohan::An attempt at getting a better ball representation. We
   * shall maintain the relative ball information (in the robot's
   * coordinate frame) as a unimodal Gaussian...
   */
  void UpdateBallGaussian();

  /**
   * Mohan::Funtion calculates the global ball position based on the
   * Gaussian estimation technique...
   */
  void UpdateGlobalBall();

  /**
   * Mohan::Function returns true if the input Gaussian represents a
   * valid point on the field - for now only a stdev check...
   */ 
  bool ValidGauss( Gauss g );

  /**
   * Mohan::Function returns true iff the ball is moving away from the
   * robot - relative motion:-)...
   */
  bool BallMovingAway();

  Point2D OutFieldPoint;
  Point2D RelBall;
  Point2D GlobBall;
  Gauss RelBallGauss;
  // heading and distance for the relball
  AngDeg RelBallHeading;
  double RelBallDistance;

  bool RecentlyStartedDegrade;
  bool SpecialVelUpdate;

  bool UsingMergedEstimate;

  // Mohan::Variables for ball velocity computation - all over
  // again...
  Point2D BallValsQueue[BallQueueMax]; // Currently stores 11 values...
  double BallTimeVals[BallQueueMax];

  bool BallVelComputed;
  int NumBallVals;
  int BallValsIdx;

  Vector2D BallRelVel;
  AngDeg BallRelVelDir;

  //double GCtime;
  double LastSignalTime;

  /**
   * Mohan::Function that attempts to calculate the ball velocity and
   * direction and if successful returns true - computation happens if
   * enough data vals stored in array...
   */
  bool ComputeRelBallVel();

  /**
   * Mohan::Function returns true if the ball is seen to be
   * approaching the robot - currently this implies that the ball
   * velocity is above a threshold and the direction of approach is
   * towards the robot...
   */
  bool BallApproaching();

  /**
   * A dummy pose to use when the actual values do not exist...
   */
  Pose DummyPose;

  /**
   * A pose that is updated based on motion alone when the robot is
   * not getting useful visual information...
   */
  Pose uncappedSmoothPose; 
  // the smooth pose from our last Brain cycle
  Pose SmoothPose;

  /**
   * A history of the (x,y)-coordinates of recent poses.  At index 0 
   * is the previous pose, index 1 the pose before that, etc.  -peggy
   */
  Point2D SmoothPoseLocationHistory[POSE_HISTORY_SIZE];

  /**
   * Returns the distance moved since the specified number of frames ago,
   * according to SmoothPoseLocationHistory.  If framesAgo is larger than 
   * POSE_HISTORY_SIZE, it will return the distance moved since 
   * POSE_HISTORY_SIZE frames ago.  The distance moved is averaged over the
   * specified number of frames ("window") after framesAgo.  -peggy
   */ 
  Distance averageDistanceMovedSince(int framesAgo, int window);

  /**
   * Update smooth pose, not letting it move far from uncappedSmoothPose
   */
  void updateSmoothPoseWithCap();

  /**
   * Nate: returns true if we are at a point p on the field within
   * some threshold
   */
  bool IsAtPoint( Point2D p, double distanceThreshold );

  /**
   * Mohan::returns true iff the pose represents a point on the
   * field...
   */
  bool IsOnField( Pose pose );

  /**
   * Mohan::Function that returns true iff the given Point2D is on
   * field...
   */
  bool IsOnField( Point2D p );

  /**
   * Mohan::Function that returns true iff the given Gauss structure
   * represents a point that is on field...
   */
  bool IsOnField( Gauss g );

  /**
   * gets heading from current position to the given point.  --peggy
   */
  AngDeg getHeadingToPoint(Point2D p);

  /**
   * Mohan::Provides the heading (wrt y-axis) to opponents' goal...
   */
  AngDeg HeadingToOppGoal();


  /**
   * Nate: the location of the opponent's goal
   */ 
  Point2D getOppGoalPoint();

  /**
   * Nate: the location of the opponent's goal
   */ 
  Point2D getBestOppGoalPoint();
  bool ChangedTargetPoint;
  Time prevpictime;
  bool manpiccap;

  /**
   * Nate: The angle in field coordinates to the opponents goal from a
   * destination point given as the argument...
   */ 
  AngDeg GlobalAngleToOppGoal( Point2D dest );

  /**
   * Mohan::Returns true if a candidate opponent goal blob has been
   * seen in vision, in the current frame...
   */
  bool OpponentGoalBlobInVision();
  
  /**
   * Mohan::Returns the observation corresponding to the opponent goal
   * colored blob that could possibly be part of the opponent goal -
   * USE WITH CARE - ASSUMES THAT A GOAL COLORED BLOB HAS BEEN SEEN IN
   * CURRENT FRAME. ALSO, THIS BLOB COULD BE NOISE...
   */
  Observation OppGoalBlobObsr();

  /**
   * Mohan::Provides a heading (wrt robot y-axis) to the candidate
   * blob that could possibly be the opponent goal - USE WITH CARE -
   * USE IFF ROBOT IS VERY CLOSE TO THE GOAL, ASSUMES THAT CANDIDATE
   * BLOB IS SEEN...
   */
  AngDeg HeadingToSeenOppGoalBlob();

  /**
   * uses vision first to determine opp goal heading, then uses
   * localization (HeadingToOppGoal()) if opp goal is not in 
   * vision.  --peggy 8mar04
   */
  AngDeg HeadingToSeenOppGoal();

  /**
   * Mohan::Gives the heading (wrt robot's y-axis) to the point to shoot
   * to when the robot tries to shoot to target but cannot find the
   * target...
   */
  AngDeg HeadingWhenTargetUnseen();

  /**
   * Mohan::Provides the heading (wrt y-axis) to own goal...
   */
  AngDeg HeadingToOwnGoal();

  /**
   * Dan::Provides the distance to own goal...
   */
  double DistanceToOwnGoal();

  /**
   * Dan::Provides the distance to opp goal...
   */
  double DistanceToOppGoal();

  /**
   * Mohan::Similar to the function for heading, the robot first tries
   * to obtain the distance to the Opponent's goal based on vision
   * objects and if that is not successful, it uses the standard
   * approach of taking the known global coordinates of the goal - USE
   * WITH CAUTION AND EVEN THEN IFF THE ROBOT IS CLOSE TO THE
   * OPPONENT'S GOAL...
   */
  double DistanceToSeenOppGoal();

  /**
   * Mohan::Function that returns the heading to the opponent left
   * corner...
   */
  AngDeg HeadingToOppLeftCorner();

  /**
   * Mohan::Function that returns the heading to the opponent left
   * corner...
   */
  AngDeg HeadingToOppRightCorner();

  /**
   * peggy: returns the heading to the midpoint of the right half
   * of the offensive end of the field (just right of the offense
   * right goal post).
   */
  AngDeg HeadingToOppMidRightSide();

  /**
   * peggy: returns the heading to the midpoint of the left half
   * of the offensive end of the field (just left of the offense
   * left goal post).
   */
  AngDeg HeadingToOppMidLeftSide();

  /**
   * Mohan::returns the heading (wrt y-axis) to the left corner of the
   * opponent's goal...
   */ 
  AngDeg HeadingToOppGoalLeftCorner();

  /**
   * Mohan::returns the heading (wrt y-axis) to the right corner of the
   * opponent's goal...
   */
  AngDeg HeadingToOppGoalRightCorner();

  /**
   * Mohan::Function that returns the heading to the center of the
   * field...
   */
  AngDeg HeadingToFieldCenter();

  /**
   * Dan: Is point in our goal box?
   */
  bool InOwnGoalBox(Point2D p);

  /**
   * Nate: are we in a place on the field where the angle to the goal
   * posts is really oblique?
   */
  bool InOffensiveWasteland();

  /**
   * Nate: heading to the point right in front of the opponent's goal
   */
  AngDeg HeadingToOffensiveKey();

  /**
   * Mohan::Provides the pose of the teammate whose ID is specified as
   * input...
   */
  Pose GetTeammatePose( int tmid );

  /**
   * Mohan::returns the pose of the teammate with specific roles...
   */
  Pose GetPoseOfAttacker();
  Pose GetPoseOfDefender();
  Pose GetPoseOfSupporter();
  Pose GetPoseOfKeeper();
  Pose GetPoseOfRobotWithRole( role_t role );

  Pose GetStartingPose();

  /**
   * Mohan::Function returns the position that the robot has to go to if
   * it decides to spread out to find the ball...
   */
  Pose GetSpreadPose();

  /**
   * Returns the smoothly varying pose...
   */
  Pose GetSmoothPose();

  /**
   * Mohan::Fucntion that returns true if the robot is within a certain
   * buffer zone around the goal box...
   */
  bool NearOwnGoalBox();

  /**
   * Selim: Function that returns true if the ball is right along the
   * opponent goal box
   */
  bool BallAtOpponentGoalLine();

  /**
   * Mohan::Fucntion that returns true if the ball is within a certain
   * buffer zone around the goal box - a modification of the function
   * below, for the keeper...
   */
  bool BallInKeeperZone();

  /**
   * Mohan::Fucntion that returns true if the ball is within a certain
   * buffer zone around the goal box...
   */
  bool BallNearOwnGoalBox();

  /**
   * Nate: like above, but for the opponent's goal box
   */
  bool BallNearOppGoalBox( double buffer = 100.0 );

  /**
   * Mohan::Function returns true if the ball is in the robot's
   * half...
   */
  bool BallInOwnHalf();

  /**
   * Mohan::Function returns true if the ball is in the opponent's
   * half...
   */
  bool BallInOpponentHalf();

  /**
   * Mohan::Function returns true if the ball is in the top third of
   * the field - depends on the side the robot is defending...
   */
  bool BallInTopThird();

  /**
   * Nate: like BallInTopThird(), but bottom instead.
   */
  bool BallInBottomThird();

  /**
   * Is the ball in the middle third of the field?
   */
  bool BallInMiddleThird();

  /**
   * Mohan::Function returns true if the ball is in the correct zone for
   * a switch to happen between the attacker and defender roles...
   */
  bool BallInSwitchZone();

  /**
   * Mohan::Function returns the x-axis displacement between the robot
   * and the ball in global coordinates...
   */
  double BallXDisplacement();
  
  /**
   * Mohan::Function returns the y-axis displacement between the robot
   * and the ball in global coordinates...
   */
  double BallYDisplacement();

  /**
   * Mohan::Function is a "weak" clone of the function above that has a
   * low confidence limit for the defender to move up...
   */
  bool BallWellInOppZone();

  /**
   * Not Mohan::returns true is ball is on green carpet
   */
  bool BallOnGrass();

  /**
   * Functions to test confidence in our position, ball position, and
   * both.
   */
  bool BallConfident();
  bool BodyConfident();
  bool BallAndBodyConfident();

  /**
   * Mohan::A version of the 'ball close to borders' function that just
   * returns true iff the ball is close to a field edge...
   */
  bool BallCloseToBorders();

  /**
   * Mohan::Variables that represent the corners and border lines of
   * the field - bot, top, llft an lrgt are numbered 0,1,2,3
   * respectively...
   */
  Point2D LL, LR, UL, UR;
  Line2D lbot, ltop, llft, lrgt;
  int LineNum;

  /**
   * Nate: returns a global Point2D of the closest point on the edge
   * of the field to the ball
   */
  Point2D ClosestEdgePointToBall();

  /**
   * Nate: Returns true if the robot is in a bad situation in which to
   * approach the ball: namely, the ball is close to a wall and the
   * robot is also close to that wall.  If we return true, we'll also
   * set turnDirection to be the lateral direction in which we should
   * move to better approach the ball.
   */ 
  //bool BadBallOnWallSituation( GenericDir& turnDirection );

  /**
   * Mohan: A variation of the function above - returns true if the
   * robot is in a bad situation in which to approach the ball: namely,
   * the ball is close to the field border - the robot should not
   * approach it head on now...
   */ 
  //bool BallCloseToFieldBorder( GenericDir& turnDirection );
  // If ball is close to the borders, this is the angle to compensate
  // for...
  AngDeg CompensAng;

  /**
   * Mohan::Function returns true if the ball is well into the offensive
   * half and its x-ccordinates are withina buffer region of the goal
   * box...
   */
  bool BallCloseToOppGoalBox();

  /**
   * Mohan::Function returns the desired ball approach angle, given that
   * it is close to a field border...
   */
  AngDeg DesiredBallApproachAng();

  /**
   * Mohan::Function returns true if the robot is in the opponent's half
   * of the field...
   */
  bool InOpponentHalf();

  /**
   * Mohan::Function returns true iff the robot is in the top third of
   * the field (team-relative coordinates)...
   */
  bool RobotInTopThird();

  /**
   * Given what team we're on, return true if we're on the right side
   * of the field
   */
  bool OnRightSideOfField();

  /**
   * Which side of the field we're on, in field coordinates: where
   * right is x > 1450
   */
  bool OnAbsoluteRightSideOfField();

  /**
   * True if we're facing the top of the field, in team coordinates:
   * top of field => our theta in [-90,90]
   */
  bool FacingTopOfField();

  /**
   * True if we're facing the top of the field, in field coordinates:
   */
  bool FacingAbsoluteTopOfField();

  /**
   * Mohan::Fucntion returns true if the robot is the closest to the
   * ball...
   */
  bool ClosestToBall();

  /**
   * Mohan::Function returns the supporting position that the robot
   * needs to go to...
   */
  Point2D GetSupportPoint();
  // Nate: a holding variable for the current support point (at least
  // for attackers, not necessarily for anyone else)
  Point2D supportPoint;
  bool resetSupportPoint;

  /**
   * Mohan::Function returns true iff the current (well known)
   * position of the ball is in the left half of the field...
   */
  bool BallInLeftHalf();

  /**
   * Mohan::Function returns true iff the current (well known)
   * position of the ball is in the rigth half of the field...
   */
  bool BallInRightHalf();

  /**
   * Mohan::Function returns true iff the robot working as the
   * defender role is in the left half of the field...
   */
  bool DefenderInLeftHalf();
  
  /**
   * Mohan::Function returns true iff the robot working as the
   * defender role is in the right half of the field...
   */
  bool DefenderInRightHalf();

  /**
   * Mohan::Function that provides the defensive support point that a
   * robot (usually the defender) needs to go to...
   */
  Point2D GetDefensivePost();

  /**
   * Mohan::Function returns true if the robot playing as the attacker
   * and actually going to ball is on the left side of the field...
   */
  bool AttackerInLeftHalf();

  /**
   * Mohan::Function returns true if the robot playing as the attacker
   * and actually going to ball is on the right side of the field...
   */
  bool AttackerInRightHalf();
  
  /**
   * Mohan::Function returns true if the robot is alone - not in
   * communication with any other robot...
   */
  bool RobotAlone();
  
  /**
   * Mohan::Function returns true if the robot is reasonable well
   * localized - certainty above a constant...
   */
  bool WellLocalized();

  /**
   * Mohan::Function returns true is the keeper is at its post...
   */
  bool KeeperInPlace();

  /**
   * Mohan::Function returns the pose that the keeper has to attain
   * under the default conditions...
   */
  Pose GetKeeperDefaultPose();

  /**
   * Mohan::Function returns the keeper position when the ball
   * position is known - a function of the ball position...
   */
  Pose GetKeeperPose();
  bool KeeperAmI;

  /**
   * Returns true if an opponent is to the left and close enough to
   * interfere with ball acquisition.  -peggy
   */
  bool OpponentImmediatelyLeft();

  /**
   * Returns true if an opponent is to the right and close enough to
   * interfere with ball acquisition.  -peggy
   */
  bool OpponentImmediatelyRight();

  /**
   * Returns true if an opponent is both closer than maxdist and in the specified heading
   * range.  -peggy  (warning...this won't work right if you expect the heading range to
   * include the discontinuity at 180/-180)
   */
  bool OpponentCloseInHeadingRange(double maxdist, double minheading, double maxheading);

  /**
   * Returns true if any opponent is within threshold of this robot.  -peggy
   */
  bool OpponentWithinDist(double threshold);

  /**
   * Mohan::Function that returns true if there exists an opponent that
   * is within buffer region of 30 degrees and a distance of 500mm from
   * the robot...
   */
  bool OpponentInWay();

  /**
   * Mohan::Function returns true if there is atleast one opponent near
   * (within a threshold buffer region around) the goal box - used for
   * keeper attack:-)...
   */
  bool OpponentNearOwnGoalBox();

  /**
   * Mohan::Function that returns the position (Point2D in global
   * coordinates) of the opponent that is the closest to the robot's own
   * goal box...
   */
  Point2D GetOppClosestToOwnGoalBox();

  /**
   * Mohan::Function returns true if the robot is within a threshold
   * distance from a field border...
   */
  bool CloseToBorders();

  /**
   * Mohan::Function returns true iff the robot cannot get to the ball -
   * say stuck because of opponents, field walls etc...
   */
  bool CannotGetToBall();


  /**
   *  Mohan::Variables for the function (below) to determine when the
   *  robot's path shall take it through the goal box region...
   */
  Point2D LineAP1, LineAP2, LineBP1, LineBP2, LineCP1, LineCP2;
  Point2D LineDP1, LineDP2, LineEP1, LineEP2, LineFP1, LineFP2;

  Line2D LineA, LineB, LineC, LineD, LineE, LineF;

  int NumGBInts;
  Point2D GBIntersects[3], ClosestInt;

  /**
   * Mohan::Function returns true iff the ball is withing the robot's
   * own goal box. Thr robot is assumed to be well localized and is
   * assumed to have a good idea of the ball...
   */
  bool BallInOwnGoalBox();

  /**
   * Mohan::Function returns true iff the current path from the robot to
   * the ball takes the robot through its own goal box - the function
   * returns false if the robot has a low confidence of its pose and/or
   * the ball position...
   */
  bool PathThroughGoalBox();

  /**
   * WARNING::FUNCTION MAKES SENSE IFF THE FUNCTION ABOVE RETURNS TRUE.
   * Mohan::Given that the current path of the robot to the ball does
   * involve walking right through its own penalty zone, this function
   * determines the "safe" point to walk to around the goal box...
   */
  Point2D NewGBTargetLocation();

  
  int BallDistID;
  double BallDistsArr[ MAX_BALL_DISTANCES ];

  // Function approximation variables...
  bool CollectData;
  double BeacDists[ 250 ];
  void AverageLandMarkDist( double dist, double angle );

  double MeasuredDistHist[ 4500 ];
  int Measdistcount;

  // Stability test calculation variable...
  bool StartStabilityTest;
  int NumArrVals;
  double PoseArr[600][3];
  void StabilityCalculations();

  int NumTrials;
  bool HistCleaned;
  Time PrevTime;


  // A variable to be used when the robot is trying to save images to
  // generate a movie file...
  bool TimeToStop;


  // Mohan::Variables for Mazda's surveillance task...
  bool markerinframe;
  ObjectT min_marker_id;
  double min_marker_offset, min_marker_angle, angleft, angright;

  
  
  /**
   * MOHAN::VARIABLES AND FUNCTIONS TO TRY AND PERFORM OPPONENT
   * DETECTION...
   */
  Observation OppObservations[ NUM_OPPONENTS ];
  int NumOpponents;
  MHGauss Opponents[ NUM_OPPONENTS ];
  MHGauss DummyOpp;
  int PrevNumOpps;
  
  Observation TmsObservations[ NUM_TEAM_MEMBERS ];
  int NumTeammates;
  MHGauss Teammates[ NUM_TEAM_MEMBERS ];
  int PrevNumTms;
  
  /**
   * Mohan::A separate function to add teammate observation - similar
   * to the one above for adding other observations of landmarks
   * etc. This makes processing of teammate observations easier...
   */
  void addTmsObservation( ObjectT obj, double dist,
			  AngDeg ang, double certainty );

  /**
   * Mohan::Function that generates an estimate of the opponents -
   * adds them one at a time to the list of opponents. It tries to
   * merge each new observation with one of the old ones. It is fails
   * it adds them to the list (if space permits) or replaces the least
   * likely estimate currently in the list...
   */
  void UpdateOpponents();

  /**
   * Mohan::Function that takes an input Gaussian estimate of an
   * opponent and tries to merge it or add it on to the already
   * existing list of known opponents...
   */
  void MergeWithOrAddOpponent( MHGauss InpOpp, bool Comm );

  /** 
   * Mohan::Function returns true if two candidate opponents are close
   * enough to be considered as the same opponent...
   */
  bool Opponents_Close( MHGauss newopp, MHGauss oldopp );

  /**
   * Mohan::Function that replaces the least likely (highest variance)
   * estimate of opponent with the latest estimate of an opponent...
   */
  void Replace_Uncertain_Opponent( MHGauss newopp, bool Comm );

  /**
   *  Mohan::Function that merges the robot's estimate of opponents
   *  with those communicated by its teammates to obtain merged
   *  estimates of opponents...
   */
  void UpdateMergedOpponents();


  // SIMILAR TO OPPONENTS, WE NEXT DO THE SAME FOR THE TEAMMATES...
  /**
   * Mohan::Function that generates an estimate of the teammates -
   * adds them one at a time to the list of teammates. It tries to
   * merge each new observation with one of the old ones. It is fails
   * it adds them to the list (if space permits) or replaces the least
   * likely estimate currently in the list...
   */
  void UpdateTeammates();

  /**
   * Mohan::Function that takes an input Gaussian estimate of a seen
   * teammate and tries to merge it or add it on to the already
   * existing list of known seen teammates...
   */
  void MergeWithOrAddTeammate( MHGauss InpTms, bool Comm );

  /** 
   * Mohan::Function returns true if two candidate seen teammates are
   * close enough to be considered as the same object...
   */
  bool Teammates_Close( MHGauss newtms, MHGauss oldtms );

  /**
   * Mohan::Function that replaces the least likely (highest variance)
   * estimate of seen teammate with the latest estimate of a
   * teammate...
   */
  void Replace_Uncertain_Teammate( MHGauss newtms, bool Comm );

  /**
   *  Mohan::Function that merges the robot's estimate of seen
   *  teammates with those communicated by its teammates to obtain
   *  merged estimates of seen teammates...
   */
  void UpdateMergedTeammates();


  // Mohan:: An attempt at trying a sector-based approach at tackling
  // opponents - a sector currently occupies 20 degrees...
  bool RelSectorsOccupied[18];

  /**
   * Mohan::Function returns true if the input angle (wrt robot's
   * x-axis) corresponds to a sector where there are no "obstacles"
   * (currently considering only opponents)...
   */
  bool SafeDirection( AngDeg relang );

  /**
   * Mohan::Function takes in the (x,y,z) point in robot-body-center
   * coordinate frame (for the z coordinate, the ground == 0) and
   * returns an overall tilt and pan in the robot-neck-centric
   * coordinate frame...
   */
  void GetAngles2Point3D( double relx, double rely, double relz, 
			  double &tilt, double &pan );
  
  double Body_Height;


  //@@@@@@@@@@@ ACCELEROMETER-BASED COLLISION DETECTION @@@@@@@@@@@
  /**
   * Mohan::Variables to be used to save and calculate the
   * accelrometer-based values for collision detection...
   */
  double XYZAccelStore[AVG_LIMIT][3];
  double CalculatedValues[TRAINING_SAMPLE_SIZE][7];
  int NumAccVals, NumCalVals;
  double accsumx, accsumy, accsumz;
  Time tslavs;

  int CollClass;

  /**
   * Mohan::Function that stores the values of the accelerometer valus
   * into the array to be used for averaging and generating training
   * samples later...
   */
  void AccelValsStore( double accelx, double accely, double accelz );

  /**
   * Mohan::Function that calculates the statistics for the current
   * sample based on the values stored in the last 1 second - generates
   * one training sample...
   */
  void GenerateStats();

  /**
   * Mohan::Function that calls the functions to first store the
   * accelerometer valus for a second and then calculate the statistics
   * to provide the training sample...
   */
  void GenerateATSamples();
  
  /**
   * Mohan::Reset accelerometer readings array index and some other
   * values if the computation is requested to be stopped...
   */
  void StopResetVals();


  // NOW ON TO TESTING ON THE ROBOT...

  // Note: Using AccelValsStore function and some of the variables
  // above to store values for testing as well - DO NOT TRY TO
  // TRAIN/TEST WITHOUT SUITABLY SETTING THE FLAG - COLLISION_DETECT -
  // DEFINED BELOW...


  /**
   * Mohan::Function stores values of Accelerometers over a few frames
   * and then computes the sample to test for collision...
   */
  void GenerateTestSample();

  double V_x, V_y, V_z, R_xy, R_yz, R_zx;
  bool NewCollVals;

  bool InCollision;
  /**
   * Mohan::Function that takes the accelrometer statistics as inputs
   * to decide if the robot is colliding or not - returns true iff the
   * robot is colliding...
   */
  bool CollidingAmI();

  /**
   * Mohan::Function takes the current velocities into account to decide
   * the new point to reach so as to avoid collisions...
   */
  Point2D AntiCollisionPoint();

  //@@@@@@@@@@@ COLOR LEARNING TASK PARAMETERS @@@@@@@@@@@
  bool RunColorLearnTask;
  int ColorLearnOption;
  int ColorClearOption;
  bool LearnGoalColorFromBeacon;
  bool GoalInView;
  int GoalIVCount;
  int ColorSpace;

  // Flag is set iff problematic colors are being learnt so that the
  // variance limits in the color map can be modified and the map can
  // be updated before learning of the colors starts...
  bool SpecialCMapUpdate;

  // Flag set iff the robot needs to search for a suitable goal blob
  // while moving around...
  bool LookForGoalBlob;

  // Parameters that are used to change the limits on the variations
  // allowed on color channels...
  bool ChangedLimits;
  int ColorChanged;
  int YLimit, CbLimit, CrLimit;

  // Mohan::Additional parameters to get the robot to focus on an
  // object in its field of view...
  double ObjectPanAngle;
  double ObjectTiltAngle;

  // Mohan::Parameters that allow the robot to learn the color better
  // by localizing to special positions on the field...
  bool ReinforceColor;
  int ReinforcePos;
  bool ReinforceCMapUpdate;

  // Mohan:: Special flag for running the color learning for
  // Dr. Lavender's demos...
  bool specialLearn;
  bool AskForBall;
  bool specialDone;
  bool SP_GWLearnt;
  bool SP_BallColLearnt;

  // Mohan::Flag to be set when the robot is performing a controlled
  // dribble...
  bool ControlDribble;

 private:

  // thresholds for each sensory input: no activation of custom senses
  // until the value exceeds this number
  static const double senseThresholds[];

  static const bool DEBUG = false;
  static const bool BALLDEBUG = false;
  static const bool COLLISION_DETECT = true;


  static const int CONSIDER_DEAD_THRESHOLD = 100;
  static const int START_PINGING_THRESHOLD = 50;
  static const int CYCLES_BETWEEN_PINGS = 25;

  /**
   * Data for handling aggregate sensory inputs: keeps track of
   * patterns of inputs, or senses  (for use in custom senses)
   */
  // the # of cycles that this sense has been lapsing
  int lapseCycles[NUM_SENSORS];
  // the # of cycles of the last activation of this sense
  int lastDuration[NUM_SENSORS];
  // the # of cycles that the sense was lapsing before this
  int prevLapseCycles[NUM_SENSORS];
  // the # of cycles of the last last activation of this sense
  int prevDuration[NUM_SENSORS];



};

#endif // WorldState_H
