#include <stdio.h>
#include <string.h>

#include <WorldState.h>
#include <PFLocalization.h>
#include <Geometric.h>

#include "Connection.h"
#include "Player.h"
#include "Parse.h"

Player::Player( WorldState *ws, Connection *c,
		PFLocalization *l )
{
  char str[MAX_MSG];

  WS = ws;
  C = c;
  L = l;

  timeLastSeeMessage = -1;
  timeLastSenseMessage = -1;
  timeLastAction = -1;
 
  bContinueLoop = true;

  sprintf( str, "(init)" );
  C->sendMessage( str );
}

void Player::mainLoop()
{
  char msg[MAX_MSG];


  while ( bContinueLoop ) {
    int n = 0;
    if ( ( n = C->receiveMessage( msg, MAX_MSG ) ) < 0 )
      break;

    if ( n > 0 ) {
      analyzeMessage( msg );
      if ( timeToAct() ) {
	L->UpdateWorldState( WS );
	WS->Update();
	L->getInfoMessage( msg );
	C->sendMessage( msg );
	WS->getInfoMessage( msg );
	C->sendMessage( msg );
	act();
	setTimeLastAction();
	WS->resetObservations();
	WS->ballInVision = false;
      }
    }
  }

  Log.log( 1, "(Player) main loop completed" );
}

void Player::act()
{
  //oneLayerHeadScan();
  figure8();
  //paramWalk (0,160,40) ; //(0, 0.0001, 0.0001);
}

void Player::figure8()
{
  static const int BUFFER = 900;
  static const int GOAL_BUFFER = 300;
  static const int NUM_TARGETS = 8;
  static const double EPSILON_D = 50.0;
  static const AngDeg EPSILON_A = 10.0;

  static const Point2D target[NUM_TARGETS] = 
    {
      Point2D( 0, 0 ),
      Point2D( -FIELD_WIDTH/2, FIELD_LENGTH/2-BUFFER),
      Point2D( 0, FIELD_LENGTH/2 - GOAL_BUFFER),
      Point2D( FIELD_WIDTH/2, FIELD_LENGTH/2-BUFFER),
      Point2D( 0, 0 ),
      Point2D( -FIELD_WIDTH/2, -FIELD_LENGTH/2+BUFFER),
      Point2D( 0, -FIELD_LENGTH/2 + GOAL_BUFFER),
      Point2D( FIELD_WIDTH/2, -FIELD_LENGTH/2+BUFFER),
    } ;
  
  static int t = 0;

  // On top of target, proceed to next one
  if ( WS->distanceToPoint( target[ t ] ) < EPSILON_D )
    t = ( t + 1 ) % NUM_TARGETS;

  // Decide body action based on angle to target
  AngDeg angle = WS->angleToPoint( target[ t ] ) - 90.0;
  if ( angle < -EPSILON_A ) {
    strafeLeft();
  } else if ( angle > EPSILON_A ) {
    strafeRight();
  } else {
    fastWalk();
  }

  // Head action: head scan
  oneLayerHeadScan();

}

void Player::paramWalk( double dx, double dy, double dt )
{
  static double vx = 0;
  static double vy = 0;
  static double vt = 0;

  char msg[MAX_MSG];

  if ( dx != vx || dy != vy || dt != vt ) {
    vx = dx; vy = dy; vt = dt;
    sprintf( msg, "(param_walk %d %d %d)", 
	     (int) dx, (int) dy, (int) dt );
    C->sendMessage( msg );
    WS->setVelocities( Vector2D( dx, dy ), dt );
  }
}

void Player::moveHead( double dp, double dt )
{
  static double vp = 0;
  static double vt = 0;

  //cerr << "moveHead(" << dp << "," << dt << ")\n";

  char msg[MAX_MSG];

  if ( dp != vp || dt != vt ) {
    vp = dp; vt = dt;
    sprintf( msg, "(move_head %d %d)", 
	     (int) dp, (int) dt );
    C->sendMessage( msg );
  }
}

void Player::strafeRight()
{
  paramWalk( 0, 0, 40 );
}

void Player::strafeLeft()
{
  paramWalk( 0, 0, -40 );
}

void Player::fastWalk()
{
  paramWalk( 0, 340, 0 );
}

void Player::oneLayerHeadScan()
{
  static const AngDeg ANGLE = 60;
  static const double EPSILON_A = 2;
  static int right = 0;
  
  if ( right ) {
    if ( WS->getHeadPan() > ANGLE - EPSILON_A )
      right = 0;
    else
      moveHead( ANGLE, 0 );
  }
  else if ( !right ) {
    if ( WS->getHeadPan() < -ANGLE + EPSILON_A )
      right = 1;
    else
      moveHead( -ANGLE, 0 );
  }
}

int Player::getCurrentTime()
{
  return timeLastSenseMessage;
}

void Player::setTimeLastSeeMessage( int t ) 
{
  timeLastSeeMessage = ( t == -1 ) ? getCurrentTime() : t;
}

void Player::setTimeLastSenseMessage( int t ) 
{
  timeLastSenseMessage = t;
  WS->time = t * 50 * MILLI;
}

void Player::setTimeLastAction( int t ) 
{
  timeLastAction = ( t == -1 ) ? getCurrentTime() : t;
}

bool Player::timeToAct() 
{
  return ( timeLastAction < getCurrentTime() );
}


void Player::processLandmarkInfo( int id, int dist, int ang, 
				  int certainty )
{
  WS->addObservation( (ObjectT) id, dist, ang, certainty * PERCENT );
}

void Player::processLinePixInfo( int id, int dist, int ang, 
				 int certainty ) {
  
}

void Player::processHeadSense( int pan, int tilt )
{
  WS->headPan = pan;
  WS->headTilt = tilt;
}

bool Player::analyzeMessage( char *msg ) 
{
  switch( msg[1] ) {
  case 'c':
    // connect - ignore
    break;
  case 's':
    switch ( msg[3] ) {
    case 'e':
      return analyzeSeeMessage( msg );
    case 'n':
      return analyzeSenseMessage( msg );
    default: break;
    }
  default:
    cerr << getCurrentTime() << ": (analyzeMessage) " 
	 << "ignored message: " << msg << endl;
    return false;
  }
}

bool Player::analyzeSeeMessage( char *msg ) 
{
  int iTime = Parse::parseFirstInt( &msg ); 
  if ( timeLastSenseMessage != iTime ) {
    cerr << ": (analyzeSeeMessage) see with different time as sense:"
         << timeLastSenseMessage << " vs. " << iTime << endl;
    return false;
  }

  setTimeLastSeeMessage( timeLastSenseMessage );

  WS->resetObservations();
  WS->ballInVision = false;
  while( *msg != ')' ) {
    msg += 2;     
  
    int id, dist, ang, cert;
    switch ( msg[0] ) {
    case 'l':
      id = Parse::parseFirstInt( &msg );
      dist = Parse::parseFirstInt( &msg );
      ang = Parse::parseFirstInt( &msg );
      cert = Parse::parseFirstInt( &msg );
      processLandmarkInfo( id, dist, ang, cert ); 
      //cout << id << ", " << dist << ", " << ang << " || ";
      break;
    case 'b':
      dist = Parse::parseFirstInt( &msg );
      ang = Parse::parseFirstInt( &msg );
      cert = Parse::parseFirstInt( &msg );
      processLandmarkInfo( OBJ_BALL, dist, ang, cert );
      //cout << "B, " << dist << ", " << ang << " || ";
      WS->ballInVision = true;
      break;
    case 'z':
      id = Parse::parseFirstInt( &msg );
      dist = Parse::parseFirstInt( &msg );
      ang = Parse::parseFirstInt( &msg );
      cert = Parse::parseFirstInt( &msg );
      processLinePixInfo( id, dist, ang, cert );
      //cout << id << "," << dist << "," << ang << " || ";
      break;
    default:
      cerr << ": (analyzeSeeMessage) unknown object type: " << msg[0] << endl;
      return false;
    }

    msg++;
  }
  //cout << endl;
  return true;
}

bool Player::analyzeSenseMessage( char *msg ) 
{
  int iTime = Parse::parseFirstInt( &msg );
  //int timeOld = getCurrentTime();
  
  while( *msg != ')' ) {
    msg += 2;     
  
    int pan, tilt;
    switch ( msg[0] ) {
    case 'h':
      pan = Parse::parseFirstInt( &msg );
      tilt = Parse::parseFirstInt( &msg );
      processHeadSense( pan, tilt );
      break;
    default:
      cerr << ": (analyzeSenseMessage) unknown sense type: " << msg[0] << endl;
      return false;
    }

    Parse::gotoFirstOccurenceOf( ')', &msg );
    msg++;
  }

  setTimeLastSenseMessage( iTime );
  Log.logWithTime( 2, " end analyzing sense" );

  return true;
}
